import math
import string
from pathlib import Path
from typing import Any, Iterable, Optional, Sequence, Type, Union

import mmh3
import numpy as np
from py_rust_stemmers import SnowballStemmer

from fastembed.common import OnnxProvider
from fastembed.common.onnx_model import OnnxOutputContext
from fastembed.common.utils import define_cache_dir
from fastembed.sparse.sparse_embedding_base import (
    SparseEmbedding,
    SparseTextEmbeddingBase,
)
from fastembed.text.onnx_text_model import OnnxTextModel, TextEmbeddingWorker
from fastembed.common.model_description import SparseModelDescription, ModelSource

supported_bm42_models: list[SparseModelDescription] = [
    SparseModelDescription(
        model="Qdrant/bm42-all-minilm-l6-v2-attentions",
        vocab_size=30522,
        description="Light sparse embedding model, which assigns an importance score to each token in the text",
        license="apache-2.0",
        size_in_GB=0.09,
        sources=ModelSource(hf="Qdrant/all_miniLM_L6_v2_with_attentions"),
        model_file="model.onnx",
        additional_files=["stopwords.txt"],
        requires_idf=True,
    ),
]

MODEL_TO_LANGUAGE = {
    "Qdrant/bm42-all-minilm-l6-v2-attentions": "english",
}


class Bm42(SparseTextEmbeddingBase, OnnxTextModel[SparseEmbedding]):
    """
    Bm42 is an extension of BM25, which tries to better evaluate importance of tokens in the documents,
    by extracting attention weights from the transformer model.

    Traditional BM25 uses a count of tokens in the document to evaluate the importance of the token,
    but this approach doesn't work well with short documents or chunks of text, as almost all tokens
    there are unique.

    BM42 addresses this issue by replacing the token count with the attention weights from the transformer model.
    This allows sparse embeddings to work well with short documents, handle rare tokens and leverage traditional NLP
    techniques like stemming and stopwords.

    WARNING: This model is expected to be used with `modifier="idf"` in the sparse vector index of Qdrant.
    """

    ONNX_OUTPUT_NAMES = ["attention_6"]

    def __init__(
        self,
        model_name: str,
        cache_dir: Optional[str] = None,
        threads: Optional[int] = None,
        providers: Optional[Sequence[OnnxProvider]] = None,
        alpha: float = 0.5,
        cuda: bool = False,
        device_ids: Optional[list[int]] = None,
        lazy_load: bool = False,
        device_id: Optional[int] = None,
        specific_model_path: Optional[str] = None,
        **kwargs: Any,
    ):
        """
        Args:
            model_name (str): The name of the model to use.
            cache_dir (str, optional): The path to the cache directory.
                                       Can be set using the `FASTEMBED_CACHE_PATH` env variable.
                                       Defaults to `fastembed_cache` in the system's temp directory.
            threads (int, optional): The number of threads single onnxruntime session can use. Defaults to None.
            providers (Optional[Sequence[OnnxProvider]], optional): The providers to use for onnxruntime.
            alpha (float, optional): Parameter, that defines the importance of the token weight in the document
                versus the importance of the token frequency in the corpus. Defaults to 0.5, based on empirical testing.
                It is recommended to only change this parameter based on training data for a specific dataset.
            cuda (bool, optional): Whether to use cuda for inference. Mutually exclusive with `providers`
                Defaults to False.
            device_ids (Optional[list[int]], optional): The list of device ids to use for data parallel processing in
                workers. Should be used with `cuda=True`, mutually exclusive with `providers`. Defaults to None.
            lazy_load (bool, optional): Whether to load the model during class initialization or on demand.
                Should be set to True when using multiple-gpu and parallel encoding. Defaults to False.
            device_id (Optional[int], optional): The device id to use for loading the model in the worker process.
            specific_model_path (Optional[str], optional): The specific path to the onnx model dir if it should be imported from somewhere else

        Raises:
            ValueError: If the model_name is not in the format <org>/<model> e.g. BAAI/bge-base-en.
        """

        super().__init__(model_name, cache_dir, threads, **kwargs)
        self.providers = providers
        self.lazy_load = lazy_load

        # List of device ids, that can be used for data parallel processing in workers
        self.device_ids = device_ids
        self.cuda = cuda

        # This device_id will be used if we need to load model in current process
        self.device_id: Optional[int] = None
        if device_id is not None:
            self.device_id = device_id
        elif self.device_ids is not None:
            self.device_id = self.device_ids[0]

        self.model_description = self._get_model_description(model_name)
        self.cache_dir = str(define_cache_dir(cache_dir))

        self._specific_model_path = specific_model_path
        self._model_dir = self.download_model(
            self.model_description,
            self.cache_dir,
            local_files_only=self._local_files_only,
            specific_model_path=self._specific_model_path,
        )

        self.invert_vocab: dict[int, str] = {}

        self.special_tokens: set[str] = set()
        self.special_tokens_ids: set[int] = set()
        self.punctuation = set(string.punctuation)
        self.stopwords = set(self._load_stopwords(self._model_dir))
        self.stemmer = SnowballStemmer(MODEL_TO_LANGUAGE[model_name])
        self.alpha = alpha

        if not self.lazy_load:
            self.load_onnx_model()

    def load_onnx_model(self) -> None:
        self._load_onnx_model(
            model_dir=self._model_dir,
            model_file=self.model_description.model_file,
            threads=self.threads,
            providers=self.providers,
            cuda=self.cuda,
            device_id=self.device_id,
        )

        for token, idx in self.tokenizer.get_vocab().items():  # type: ignore[union-attr]
            self.invert_vocab[idx] = token
        self.special_tokens = set(self.special_token_to_id.keys())
        self.special_tokens_ids = set(self.special_token_to_id.values())
        self.stopwords = set(self._load_stopwords(self._model_dir))

    def _filter_pair_tokens(self, tokens: list[tuple[str, Any]]) -> list[tuple[str, Any]]:
        result: list[tuple[str, Any]] = []
        for token, value in tokens:
            if token in self.stopwords or token in self.punctuation:
                continue
            result.append((token, value))
        return result

    def _stem_pair_tokens(self, tokens: list[tuple[str, Any]]) -> list[tuple[str, Any]]:
        result: list[tuple[str, Any]] = []
        for token, value in tokens:
            processed_token = self.stemmer.stem_word(token)
            result.append((processed_token, value))
        return result

    @classmethod
    def _aggregate_weights(
        cls, tokens: list[tuple[str, list[int]]], weights: list[float]
    ) -> list[tuple[str, float]]:
        result: list[tuple[str, float]] = []
        for token, idxs in tokens:
            sum_weight = sum(weights[idx] for idx in idxs)
            result.append((token, sum_weight))
        return result

    def _reconstruct_bpe(
        self, bpe_tokens: Iterable[tuple[int, str]]
    ) -> list[tuple[str, list[int]]]:
        result: list[tuple[str, list[int]]] = []
        acc: str = ""
        acc_idx: list[int] = []

        continuing_subword_prefix = self.tokenizer.model.continuing_subword_prefix  # type: ignore[union-attr]
        continuing_subword_prefix_len = len(continuing_subword_prefix)

        for idx, token in bpe_tokens:
            if token in self.special_tokens:
                continue

            if token.startswith(continuing_subword_prefix):
                acc += token[continuing_subword_prefix_len:]
                acc_idx.append(idx)
            else:
                if acc:
                    result.append((acc, acc_idx))
                    acc_idx = []
                acc = token
                acc_idx.append(idx)

        if acc:
            result.append((acc, acc_idx))

        return result

    def _rescore_vector(self, vector: dict[str, float]) -> dict[int, float]:
        """
        Orders all tokens in the vector by their importance and generates a new score based on the importance order.
        So that the scoring doesn't depend on absolute values assigned by the model, but on the relative importance.
        """

        new_vector: dict[int, float] = {}

        for token, value in vector.items():
            token_id = abs(mmh3.hash(token))
            # Examples:
            # Num 0: Log(1/1 + 1) = 0.6931471805599453
            # Num 1: Log(1/2 + 1) = 0.4054651081081644
            # Num 2: Log(1/3 + 1) = 0.28768207245178085
            new_vector[token_id] = math.log(1.0 + value) ** self.alpha  # value

        return new_vector

    def _post_process_onnx_output(
        self, output: OnnxOutputContext, **kwargs: Any
    ) -> Iterable[SparseEmbedding]:
        if output.input_ids is None:
            raise ValueError("input_ids must be provided for document post-processing")

        token_ids_batch = output.input_ids.astype(int)

        # attention_value shape: (batch_size, num_heads, num_tokens, num_tokens)
        pooled_attention = np.mean(output.model_output[:, :, 0], axis=1) * output.attention_mask

        for document_token_ids, attention_value in zip(token_ids_batch, pooled_attention):
            document_tokens_with_ids = (
                (idx, self.invert_vocab[token_id])
                for idx, token_id in enumerate(document_token_ids)
            )

            reconstructed = self._reconstruct_bpe(document_tokens_with_ids)

            filtered = self._filter_pair_tokens(reconstructed)

            stemmed = self._stem_pair_tokens(filtered)

            weighted = self._aggregate_weights(stemmed, attention_value)

            max_token_weight: dict[str, float] = {}

            for token, weight in weighted:
                max_token_weight[token] = max(max_token_weight.get(token, 0), weight)

            rescored = self._rescore_vector(max_token_weight)

            yield SparseEmbedding.from_dict(rescored)

    @classmethod
    def _list_supported_models(cls) -> list[SparseModelDescription]:
        """Lists the supported models.

        Returns:
            list[SparseModelDescription]: A list of SparseModelDescription objects containing the model information.
        """
        return supported_bm42_models

    @classmethod
    def _load_stopwords(cls, model_dir: Path) -> list[str]:
        stopwords_path = model_dir / "stopwords.txt"
        if not stopwords_path.exists():
            return []

        with open(stopwords_path, "r") as f:
            return f.read().splitlines()

    def embed(
        self,
        documents: Union[str, Iterable[str]],
        batch_size: int = 256,
        parallel: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterable[SparseEmbedding]:
        """
        Encode a list of documents into list of embeddings.
        We use mean pooling with attention so that the model can handle variable-length inputs.

        Args:
            documents: Iterator of documents or single document to embed
            batch_size: Batch size for encoding -- higher values will use more memory, but be faster
            parallel:
                If > 1, data-parallel encoding will be used, recommended for offline encoding of large datasets.
                If 0, use all available cores.
                If None, don't use data-parallel processing, use default onnxruntime threading instead.

        Returns:
            List of embeddings, one per document
        """
        yield from self._embed_documents(
            model_name=self.model_name,
            cache_dir=str(self.cache_dir),
            documents=documents,
            batch_size=batch_size,
            parallel=parallel,
            providers=self.providers,
            cuda=self.cuda,
            device_ids=self.device_ids,
            alpha=self.alpha,
            local_files_only=self._local_files_only,
            specific_model_path=self._specific_model_path,
        )

    @classmethod
    def _query_rehash(cls, tokens: Iterable[str]) -> dict[int, float]:
        result: dict[int, float] = {}
        for token in tokens:
            token_id = abs(mmh3.hash(token))
            result[token_id] = 1.0
        return result

    def query_embed(
        self, query: Union[str, Iterable[str]], **kwargs: Any
    ) -> Iterable[SparseEmbedding]:
        """
        To emulate BM25 behaviour, we don't need to use smart weights in the query, and
        it's enough to just hash the tokens and assign a weight of 1.0 to them.
        It is also faster, as we don't need to run the model for the query.
        """
        if isinstance(query, str):
            query = [query]

        if not hasattr(self, "model") or self.model is None:
            self.load_onnx_model()

        for text in query:
            encoded = self.tokenizer.encode(text)  # type: ignore[union-attr]
            document_tokens_with_ids = enumerate(encoded.tokens)
            reconstructed = self._reconstruct_bpe(document_tokens_with_ids)
            filtered = self._filter_pair_tokens(reconstructed)
            stemmed = self._stem_pair_tokens(filtered)

            yield SparseEmbedding.from_dict(self._query_rehash(token for token, _ in stemmed))

    @classmethod
    def _get_worker_class(cls) -> Type[TextEmbeddingWorker[SparseEmbedding]]:
        return Bm42TextEmbeddingWorker


class Bm42TextEmbeddingWorker(TextEmbeddingWorker[SparseEmbedding]):
    def init_embedding(self, model_name: str, cache_dir: str, **kwargs: Any) -> Bm42:
        return Bm42(
            model_name=model_name,
            cache_dir=cache_dir,
            **kwargs,
        )



from enum import Enum
from typing import Any, Type, Iterable, Union, Optional

import numpy as np

from fastembed.common.onnx_model import OnnxOutputContext
from fastembed.common.types import NumpyArray
from fastembed.text.pooled_normalized_embedding import PooledNormalizedEmbedding
from fastembed.text.onnx_embedding import OnnxTextEmbeddingWorker
from fastembed.common.model_description import DenseModelDescription, ModelSource

supported_multitask_models: list[DenseModelDescription] = [
    DenseModelDescription(
        model="jinaai/jina-embeddings-v3",
        dim=1024,
        tasks={
            "retrieval.query": 0,
            "retrieval.passage": 1,
            "separation": 2,
            "classification": 3,
            "text-matching": 4,
        },
        description=(
            "Multi-task unimodal (text) embedding model, multi-lingual (~100), "
            "1024 tokens truncation, and 8192 sequence length. Prefixes for queries/documents: not necessary, 2024 year."
        ),
        license="cc-by-nc-4.0",
        size_in_GB=2.29,
        sources=ModelSource(hf="jinaai/jina-embeddings-v3"),
        model_file="onnx/model.onnx",
        additional_files=["onnx/model.onnx_data"],
    ),
]


class Task(int, Enum):
    RETRIEVAL_QUERY = 0
    RETRIEVAL_PASSAGE = 1
    SEPARATION = 2
    CLASSIFICATION = 3
    TEXT_MATCHING = 4


class JinaEmbeddingV3(PooledNormalizedEmbedding):
    PASSAGE_TASK = Task.RETRIEVAL_PASSAGE
    QUERY_TASK = Task.RETRIEVAL_QUERY

    def __init__(self, *args: Any, task_id: Optional[int] = None, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.default_task_id: Union[Task, int] = (
            task_id if task_id is not None else self.PASSAGE_TASK
        )

    @classmethod
    def _get_worker_class(cls) -> Type[OnnxTextEmbeddingWorker]:
        return JinaEmbeddingV3Worker

    @classmethod
    def _list_supported_models(cls) -> list[DenseModelDescription]:
        return supported_multitask_models

    def _preprocess_onnx_input(
        self,
        onnx_input: dict[str, NumpyArray],
        task_id: Optional[Union[int, Task]] = None,
        **kwargs: Any,
    ) -> dict[str, NumpyArray]:
        if task_id is None:
            raise ValueError(f"task_id must be provided for JinaEmbeddingV3, got <{task_id}>")
        onnx_input["task_id"] = np.array(task_id, dtype=np.int64)
        return onnx_input

    def embed(
        self,
        documents: Union[str, Iterable[str]],
        batch_size: int = 256,
        parallel: Optional[int] = None,
        task_id: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterable[NumpyArray]:
        task_id = (
            task_id if task_id is not None else self.default_task_id
        )  # required for multiprocessing
        yield from super().embed(documents, batch_size, parallel, task_id=task_id, **kwargs)

    def query_embed(self, query: Union[str, Iterable[str]], **kwargs: Any) -> Iterable[NumpyArray]:
        yield from super().embed(query, task_id=self.QUERY_TASK, **kwargs)

    def passage_embed(self, texts: Iterable[str], **kwargs: Any) -> Iterable[NumpyArray]:
        yield from super().embed(texts, task_id=self.PASSAGE_TASK, **kwargs)


class JinaEmbeddingV3Worker(OnnxTextEmbeddingWorker):
    def init_embedding(
        self,
        model_name: str,
        cache_dir: str,
        **kwargs: Any,
    ) -> JinaEmbeddingV3:
        return JinaEmbeddingV3(
            model_name=model_name,
            cache_dir=cache_dir,
            threads=1,
            **kwargs,
        )

    def process(self, items: Iterable[tuple[int, Any]]) -> Iterable[tuple[int, OnnxOutputContext]]:
        self.model: JinaEmbeddingV3  # mypy complaints `self.model` does not have `default_task_id`
        for idx, batch in items:
            onnx_output = self.model.onnx_embed(batch, task_id=self.model.default_task_id)
            yield idx, onnx_output



from dataclasses import dataclass
from typing import Iterable, Optional, Union, Any

import numpy as np
from numpy.typing import NDArray

from fastembed.common.model_description import SparseModelDescription
from fastembed.common.types import NumpyArray
from fastembed.common.model_management import ModelManagement


@dataclass
class SparseEmbedding:
    values: NumpyArray
    indices: Union[NDArray[np.int64], NDArray[np.int32]]

    def as_object(self) -> dict[str, NumpyArray]:
        return {
            "values": self.values,
            "indices": self.indices,
        }

    def as_dict(self) -> dict[int, float]:
        return {int(i): float(v) for i, v in zip(self.indices, self.values)}  # type: ignore

    @classmethod
    def from_dict(cls, data: dict[int, float]) -> "SparseEmbedding":
        if len(data) == 0:
            return cls(values=np.array([]), indices=np.array([]))
        indices, values = zip(*data.items())
        return cls(values=np.array(values), indices=np.array(indices))


class SparseTextEmbeddingBase(ModelManagement[SparseModelDescription]):
    def __init__(
        self,
        model_name: str,
        cache_dir: Optional[str] = None,
        threads: Optional[int] = None,
        **kwargs: Any,
    ):
        self.model_name = model_name
        self.cache_dir = cache_dir
        self.threads = threads
        self._local_files_only = kwargs.pop("local_files_only", False)

    def embed(
        self,
        documents: Union[str, Iterable[str]],
        batch_size: int = 256,
        parallel: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterable[SparseEmbedding]:
        raise NotImplementedError()

    def passage_embed(self, texts: Iterable[str], **kwargs: Any) -> Iterable[SparseEmbedding]:
        """
        Embeds a list of text passages into a list of embeddings.

        Args:
            texts (Iterable[str]): The list of texts to embed.
            **kwargs: Additional keyword argument to pass to the embed method.

        Yields:
            Iterable[SparseEmbedding]: The sparse embeddings.
        """

        # This is model-specific, so that different models can have specialized implementations
        yield from self.embed(texts, **kwargs)

    def query_embed(
        self, query: Union[str, Iterable[str]], **kwargs: Any
    ) -> Iterable[SparseEmbedding]:
        """
        Embeds queries

        Args:
            query (Union[str, Iterable[str]]): The query to embed, or an iterable e.g. list of queries.

        Returns:
            Iterable[SparseEmbedding]: The sparse embeddings.
        """

        # This is model-specific, so that different models can have specialized implementations
        if isinstance(query, str):
            yield from self.embed([query], **kwargs)
        else:
            yield from self.embed(query, **kwargs)



from typing import Union

import numpy as np
from PIL import Image

from fastembed.common.types import NumpyArray


def convert_to_rgb(image: Image.Image) -> Image.Image:
    if image.mode == "RGB":
        return image

    image = image.convert("RGB")
    return image


def center_crop(
    image: Union[Image.Image, NumpyArray],
    size: tuple[int, int],
) -> NumpyArray:
    if isinstance(image, np.ndarray):
        _, orig_height, orig_width = image.shape
    else:
        orig_height, orig_width = image.height, image.width
        # (H, W, C) -> (C, H, W)
        image = np.array(image).transpose((2, 0, 1))

    crop_height, crop_width = size

    # left upper corner (0, 0)
    top = (orig_height - crop_height) // 2
    bottom = top + crop_height
    left = (orig_width - crop_width) // 2
    right = left + crop_width

    # Check if cropped area is within image boundaries
    if top >= 0 and bottom <= orig_height and left >= 0 and right <= orig_width:
        image = image[..., top:bottom, left:right]
        return image

    # Padding with zeros
    new_height = max(crop_height, orig_height)
    new_width = max(crop_width, orig_width)
    new_shape = image.shape[:-2] + (new_height, new_width)
    new_image = np.zeros_like(image, shape=new_shape, dtype=np.float32)

    top_pad = (new_height - orig_height) // 2
    bottom_pad = top_pad + orig_height
    left_pad = (new_width - orig_width) // 2
    right_pad = left_pad + orig_width
    new_image[..., top_pad:bottom_pad, left_pad:right_pad] = image

    top += top_pad
    bottom += top_pad
    left += left_pad
    right += left_pad

    new_image = new_image[
        ..., max(0, top) : min(new_height, bottom), max(0, left) : min(new_width, right)
    ]

    return new_image


def normalize(
    image: NumpyArray,
    mean: Union[float, list[float]],
    std: Union[float, list[float]],
) -> NumpyArray:
    num_channels = image.shape[1] if len(image.shape) == 4 else image.shape[0]

    if not np.issubdtype(image.dtype, np.floating):
        image = image.astype(np.float32)

    mean_list = mean if isinstance(mean, list) else [mean] * num_channels

    if len(mean_list) != num_channels:
        raise ValueError(
            f"mean must have the same number of channels as the image, image has {num_channels} channels, got "
            f"{len(mean_list)}"
        )

    mean_arr = np.array(mean_list, dtype=np.float32)

    std_list = std if isinstance(std, list) else [std] * num_channels
    if len(std_list) != num_channels:
        raise ValueError(
            f"std must have the same number of channels as the image, image has {num_channels} channels, got {len(std_list)}"
        )

    std_arr = np.array(std_list, dtype=np.float32)

    image_upd = ((image.T - mean_arr) / std_arr).T
    return image_upd


def resize(
    image: Image.Image,
    size: Union[int, tuple[int, int]],
    resample: Union[int, Image.Resampling] = Image.Resampling.BILINEAR,
) -> Image.Image:
    if isinstance(size, tuple):
        return image.resize(size, resample)

    height, width = image.height, image.width
    short, long = (width, height) if width <= height else (height, width)

    new_short, new_long = size, int(size * long / short)
    if width <= height:
        new_size = (new_short, new_long)
    else:
        new_size = (new_long, new_short)
    return image.resize(new_size, resample)


def rescale(image: NumpyArray, scale: float, dtype: type = np.float32) -> NumpyArray:
    return (image * scale).astype(dtype)


def pil2ndarray(image: Union[Image.Image, NumpyArray]) -> NumpyArray:
    if isinstance(image, Image.Image):
        return np.asarray(image).transpose((2, 0, 1))
    return image


def pad2square(
    image: Image.Image,
    size: int,
    fill_color: Union[str, int, tuple[int, ...]] = 0,
) -> Image.Image:
    height, width = image.height, image.width

    left, right = 0, width
    top, bottom = 0, height

    crop_required = False
    if width > size:
        left = (width - size) // 2
        right = left + size
        crop_required = True

    if height > size:
        top = (height - size) // 2
        bottom = top + size
        crop_required = True

    new_image = Image.new(mode="RGB", size=(size, size), color=fill_color)
    new_image.paste(image.crop((left, top, right, bottom)) if crop_required else image)
    return new_image



from typing import Any, Iterable, Optional, Sequence, Type, Union
from dataclasses import asdict

from fastembed.common.types import NumpyArray
from fastembed.common import ImageInput, OnnxProvider
from fastembed.image.image_embedding_base import ImageEmbeddingBase
from fastembed.image.onnx_embedding import OnnxImageEmbedding
from fastembed.common.model_description import DenseModelDescription


class ImageEmbedding(ImageEmbeddingBase):
    EMBEDDINGS_REGISTRY: list[Type[ImageEmbeddingBase]] = [OnnxImageEmbedding]

    @classmethod
    def list_supported_models(cls) -> list[dict[str, Any]]:
        """
        Lists the supported models.

        Returns:
            list[dict[str, Any]]: A list of dictionaries containing the model information.

            Example:
                ```
                [
                    {
                        "model": "Qdrant/clip-ViT-B-32-vision",
                        "dim": 512,
                        "description": "CLIP vision encoder based on ViT-B/32",
                        "license": "mit",
                        "size_in_GB": 0.33,
                        "sources": {
                            "hf": "Qdrant/clip-ViT-B-32-vision",
                        },
                        "model_file": "model.onnx",
                    }
                ]
                ```
        """
        return [asdict(model) for model in cls._list_supported_models()]

    @classmethod
    def _list_supported_models(cls) -> list[DenseModelDescription]:
        result: list[DenseModelDescription] = []
        for embedding in cls.EMBEDDINGS_REGISTRY:
            result.extend(embedding._list_supported_models())
        return result

    def __init__(
        self,
        model_name: str,
        cache_dir: Optional[str] = None,
        threads: Optional[int] = None,
        providers: Optional[Sequence[OnnxProvider]] = None,
        cuda: bool = False,
        device_ids: Optional[list[int]] = None,
        lazy_load: bool = False,
        **kwargs: Any,
    ):
        super().__init__(model_name, cache_dir, threads, **kwargs)
        for EMBEDDING_MODEL_TYPE in self.EMBEDDINGS_REGISTRY:
            supported_models = EMBEDDING_MODEL_TYPE._list_supported_models()
            if any(model_name.lower() == model.model.lower() for model in supported_models):
                self.model = EMBEDDING_MODEL_TYPE(
                    model_name,
                    cache_dir,
                    threads=threads,
                    providers=providers,
                    cuda=cuda,
                    device_ids=device_ids,
                    lazy_load=lazy_load,
                    **kwargs,
                )
                return

        raise ValueError(
            f"Model {model_name} is not supported in ImageEmbedding."
            "Please check the supported models using `ImageEmbedding.list_supported_models()`"
        )

    @property
    def embedding_size(self) -> int:
        """Get the embedding size of the current model"""
        if self._embedding_size is None:
            self._embedding_size = self.get_embedding_size(self.model_name)
        return self._embedding_size

    @classmethod
    def get_embedding_size(cls, model_name: str) -> int:
        """Get the embedding size of the passed model

        Args:
            model_name (str): The name of the model to get embedding size for.

        Returns:
            int: The size of the embedding.

        Raises:
            ValueError: If the model name is not found in the supported models.
        """
        descriptions = cls._list_supported_models()
        embedding_size: Optional[int] = None
        for description in descriptions:
            if description.model.lower() == model_name.lower():
                embedding_size = description.dim
                break
        if embedding_size is None:
            model_names = [description.model for description in descriptions]
            raise ValueError(
                f"Embedding size for model {model_name} was None. "
                f"Available model names: {model_names}"
            )
        return embedding_size

    def embed(
        self,
        images: Union[ImageInput, Iterable[ImageInput]],
        batch_size: int = 16,
        parallel: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterable[NumpyArray]:
        """
        Encode a list of images into list of embeddings.

        Args:
            images: Iterator of image paths or single image path to embed
            batch_size: Batch size for encoding -- higher values will use more memory, but be faster
            parallel:
                If > 1, data-parallel encoding will be used, recommended for offline encoding of large datasets.
                If 0, use all available cores.
                If None, don't use data-parallel processing, use default onnxruntime threading instead.

        Returns:
            List of embeddings, one per document
        """
        yield from self.model.embed(images, batch_size, parallel, **kwargs)



from typing import Any, Iterable, Optional, Sequence, Type, Union

from fastembed.common.types import NumpyArray, OnnxProvider
import numpy as np

from fastembed.common.onnx_model import OnnxOutputContext
from fastembed.common.utils import define_cache_dir, normalize
from fastembed.text.onnx_text_model import OnnxTextModel, TextEmbeddingWorker
from fastembed.text.text_embedding_base import TextEmbeddingBase
from fastembed.common.model_description import DenseModelDescription, ModelSource

supported_onnx_models: list[DenseModelDescription] = [
    DenseModelDescription(
        model="BAAI/bge-base-en",
        dim=768,
        description=(
            "Text embeddings, Unimodal (text), English, 512 input tokens truncation, "
            "Prefixes for queries/documents: necessary, 2023 year."
        ),
        license="mit",
        size_in_GB=0.42,
        sources=ModelSource(
            hf="Qdrant/fast-bge-base-en",
            url="https://storage.googleapis.com/qdrant-fastembed/fast-bge-base-en.tar.gz",
            _deprecated_tar_struct=True,
        ),
        model_file="model_optimized.onnx",
    ),
    DenseModelDescription(
        model="BAAI/bge-base-en-v1.5",
        dim=768,
        description=(
            "Text embeddings, Unimodal (text), English, 512 input tokens truncation, "
            "Prefixes for queries/documents: not so necessary, 2023 year."
        ),
        license="mit",
        size_in_GB=0.21,
        sources=ModelSource(
            hf="qdrant/bge-base-en-v1.5-onnx-q",
            url="https://storage.googleapis.com/qdrant-fastembed/fast-bge-base-en-v1.5.tar.gz",
            _deprecated_tar_struct=True,
        ),
        model_file="model_optimized.onnx",
    ),
    DenseModelDescription(
        model="BAAI/bge-large-en-v1.5",
        dim=1024,
        description=(
            "Text embeddings, Unimodal (text), English, 512 input tokens truncation, "
            "Prefixes for queries/documents: not so necessary, 2023 year."
        ),
        license="mit",
        size_in_GB=1.20,
        sources=ModelSource(hf="qdrant/bge-large-en-v1.5-onnx"),
        model_file="model.onnx",
    ),
    DenseModelDescription(
        model="BAAI/bge-small-en",
        dim=384,
        description=(
            "Text embeddings, Unimodal (text), English, 512 input tokens truncation, "
            "Prefixes for queries/documents: necessary, 2023 year."
        ),
        license="mit",
        size_in_GB=0.13,
        sources=ModelSource(
            hf="Qdrant/bge-small-en",
            url="https://storage.googleapis.com/qdrant-fastembed/BAAI-bge-small-en.tar.gz",
            _deprecated_tar_struct=True,
        ),
        model_file="model_optimized.onnx",
    ),
    DenseModelDescription(
        model="BAAI/bge-small-en-v1.5",
        dim=384,
        description=(
            "Text embeddings, Unimodal (text), English, 512 input tokens truncation, "
            "Prefixes for queries/documents: not so necessary, 2023 year."
        ),
        license="mit",
        size_in_GB=0.067,
        sources=ModelSource(hf="qdrant/bge-small-en-v1.5-onnx-q"),
        model_file="model_optimized.onnx",
    ),
    DenseModelDescription(
        model="BAAI/bge-small-zh-v1.5",
        dim=512,
        description=(
            "Text embeddings, Unimodal (text), Chinese, 512 input tokens truncation, "
            "Prefixes for queries/documents: not so necessary, 2023 year."
        ),
        license="mit",
        size_in_GB=0.09,
        sources=ModelSource(
            hf="Qdrant/bge-small-zh-v1.5",
            url="https://storage.googleapis.com/qdrant-fastembed/fast-bge-small-zh-v1.5.tar.gz",
            _deprecated_tar_struct=True,
        ),
        model_file="model_optimized.onnx",
    ),
    DenseModelDescription(
        model="mixedbread-ai/mxbai-embed-large-v1",
        dim=1024,
        description=(
            "Text embeddings, Unimodal (text), English, 512 input tokens truncation, "
            "Prefixes for queries/documents: necessary, 2024 year."
        ),
        license="apache-2.0",
        size_in_GB=0.64,
        sources=ModelSource(hf="mixedbread-ai/mxbai-embed-large-v1"),
        model_file="onnx/model.onnx",
    ),
    DenseModelDescription(
        model="snowflake/snowflake-arctic-embed-xs",
        dim=384,
        description=(
            "Text embeddings, Unimodal (text), English, 512 input tokens truncation, "
            "Prefixes for queries/documents: necessary, 2024 year."
        ),
        license="apache-2.0",
        size_in_GB=0.09,
        sources=ModelSource(hf="snowflake/snowflake-arctic-embed-xs"),
        model_file="onnx/model.onnx",
    ),
    DenseModelDescription(
        model="snowflake/snowflake-arctic-embed-s",
        dim=384,
        description=(
            "Text embeddings, Unimodal (text), English, 512 input tokens truncation, "
            "Prefixes for queries/documents: necessary, 2024 year."
        ),
        license="apache-2.0",
        size_in_GB=0.13,
        sources=ModelSource(hf="snowflake/snowflake-arctic-embed-s"),
        model_file="onnx/model.onnx",
    ),
    DenseModelDescription(
        model="snowflake/snowflake-arctic-embed-m",
        dim=768,
        description=(
            "Text embeddings, Unimodal (text), English, 512 input tokens truncation, "
            "Prefixes for queries/documents: necessary, 2024 year."
        ),
        license="apache-2.0",
        size_in_GB=0.43,
        sources=ModelSource(hf="Snowflake/snowflake-arctic-embed-m"),
        model_file="onnx/model.onnx",
    ),
    DenseModelDescription(
        model="snowflake/snowflake-arctic-embed-m-long",
        dim=768,
        description=(
            "Text embeddings, Unimodal (text), English, 2048 input tokens truncation, "
            "Prefixes for queries/documents: necessary, 2024 year."
        ),
        license="apache-2.0",
        size_in_GB=0.54,
        sources=ModelSource(hf="snowflake/snowflake-arctic-embed-m-long"),
        model_file="onnx/model.onnx",
    ),
    DenseModelDescription(
        model="snowflake/snowflake-arctic-embed-l",
        dim=1024,
        description=(
            "Text embeddings, Unimodal (text), English, 512 input tokens truncation, "
            "Prefixes for queries/documents: necessary, 2024 year."
        ),
        license="apache-2.0",
        size_in_GB=1.02,
        sources=ModelSource(hf="snowflake/snowflake-arctic-embed-l"),
        model_file="onnx/model.onnx",
    ),
    DenseModelDescription(
        model="jinaai/jina-clip-v1",
        dim=768,
        description=(
            "Text embeddings, Multimodal (text&image), English, Prefixes for queries/documents: "
            "not necessary, 2024 year"
        ),
        license="apache-2.0",
        size_in_GB=0.55,
        sources=ModelSource(hf="jinaai/jina-clip-v1"),
        model_file="onnx/text_model.onnx",
    ),
]


class OnnxTextEmbedding(TextEmbeddingBase, OnnxTextModel[NumpyArray]):
    """Implementation of the Flag Embedding model."""

    @classmethod
    def _list_supported_models(cls) -> list[DenseModelDescription]:
        """
        Lists the supported models.

        Returns:
            list[DenseModelDescription]: A list of DenseModelDescription objects containing the model information.
        """
        return supported_onnx_models

    def __init__(
        self,
        model_name: str = "BAAI/bge-small-en-v1.5",
        cache_dir: Optional[str] = None,
        threads: Optional[int] = None,
        providers: Optional[Sequence[OnnxProvider]] = None,
        cuda: bool = False,
        device_ids: Optional[list[int]] = None,
        lazy_load: bool = False,
        device_id: Optional[int] = None,
        specific_model_path: Optional[str] = None,
        **kwargs: Any,
    ):
        """
        Args:
            model_name (str): The name of the model to use.
            cache_dir (str, optional): The path to the cache directory.
                                       Can be set using the `FASTEMBED_CACHE_PATH` env variable.
                                       Defaults to `fastembed_cache` in the system's temp directory.
            threads (int, optional): The number of threads single onnxruntime session can use. Defaults to None.
            providers (Optional[Sequence[OnnxProvider]], optional): The list of onnxruntime providers to use.
                Mutually exclusive with the `cuda` and `device_ids` arguments. Defaults to None.
            cuda (bool, optional): Whether to use cuda for inference. Mutually exclusive with `providers`
                Defaults to False.
            device_ids (Optional[list[int]], optional): The list of device ids to use for data parallel processing in
                workers. Should be used with `cuda=True`, mutually exclusive with `providers`. Defaults to None.
            lazy_load (bool, optional): Whether to load the model during class initialization or on demand.
                Should be set to True when using multiple-gpu and parallel encoding. Defaults to False.
            device_id (Optional[int], optional): The device id to use for loading the model in the worker process.
            specific_model_path (Optional[str], optional): The specific path to the onnx model dir if it should be imported from somewhere else

        Raises:
            ValueError: If the model_name is not in the format <org>/<model> e.g. BAAI/bge-base-en.
        """
        super().__init__(model_name, cache_dir, threads, **kwargs)
        self.providers = providers
        self.lazy_load = lazy_load

        # List of device ids, that can be used for data parallel processing in workers
        self.device_ids = device_ids
        self.cuda = cuda

        # This device_id will be used if we need to load model in current process
        self.device_id: Optional[int] = None
        if device_id is not None:
            self.device_id = device_id
        elif self.device_ids is not None:
            self.device_id = self.device_ids[0]

        self.model_description = self._get_model_description(model_name)
        self.cache_dir = str(define_cache_dir(cache_dir))
        self._specific_model_path = specific_model_path
        self._model_dir = self.download_model(
            self.model_description,
            self.cache_dir,
            local_files_only=self._local_files_only,
            specific_model_path=self._specific_model_path,
        )

        if not self.lazy_load:
            self.load_onnx_model()

    def embed(
        self,
        documents: Union[str, Iterable[str]],
        batch_size: int = 256,
        parallel: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterable[NumpyArray]:
        """
        Encode a list of documents into list of embeddings.
        We use mean pooling with attention so that the model can handle variable-length inputs.

        Args:
            documents: Iterator of documents or single document to embed
            batch_size: Batch size for encoding -- higher values will use more memory, but be faster
            parallel:
                If > 1, data-parallel encoding will be used, recommended for offline encoding of large datasets.
                If 0, use all available cores.
                If None, don't use data-parallel processing, use default onnxruntime threading instead.

        Returns:
            List of embeddings, one per document
        """
        yield from self._embed_documents(
            model_name=self.model_name,
            cache_dir=str(self.cache_dir),
            documents=documents,
            batch_size=batch_size,
            parallel=parallel,
            providers=self.providers,
            cuda=self.cuda,
            device_ids=self.device_ids,
            local_files_only=self._local_files_only,
            specific_model_path=self._specific_model_path,
            **kwargs,
        )

    @classmethod
    def _get_worker_class(cls) -> Type["TextEmbeddingWorker[NumpyArray]"]:
        return OnnxTextEmbeddingWorker

    def _preprocess_onnx_input(
        self, onnx_input: dict[str, NumpyArray], **kwargs: Any
    ) -> dict[str, NumpyArray]:
        """
        Preprocess the onnx input.
        """
        if self.model_name == "qwen3-embedding-0.6b-onnx" and "position_ids" not in onnx_input:
            # Derive positions from attention_mask to handle left/right padding correctly
            mask = onnx_input["attention_mask"].astype(np.int64, copy=False)
            pos = np.cumsum(mask, axis=1) - 1  # [-1 for pads, 0..L-1 for tokens]
            np.maximum(pos, 0, out=pos)  # clamp negatives
            pos *= mask  # ensure pads are 0, not repeated last index
            pos = np.ascontiguousarray(pos, dtype=np.int64)

            onnx_input["position_ids"] = pos
        return onnx_input

    def _post_process_onnx_output(
        self, output: OnnxOutputContext, **kwargs: Any
    ) -> Iterable[NumpyArray]:
        embeddings = output.model_output

        if embeddings.ndim == 3:  # (batch_size, seq_len, embedding_dim)
            processed_embeddings = embeddings[:, 0]
        elif embeddings.ndim == 2:  # (batch_size, embedding_dim)
            processed_embeddings = embeddings
        else:
            raise ValueError(f"Unsupported embedding shape: {embeddings.shape}")
        return normalize(processed_embeddings)

    def load_onnx_model(self) -> None:
        self._load_onnx_model(
            model_dir=self._model_dir,
            model_file=self.model_description.model_file,
            threads=self.threads,
            providers=self.providers,
            cuda=self.cuda,
            device_id=self.device_id,
        )


class OnnxTextEmbeddingWorker(TextEmbeddingWorker[NumpyArray]):
    def init_embedding(
        self,
        model_name: str,
        cache_dir: str,
        **kwargs: Any,
    ) -> OnnxTextEmbedding:
        return OnnxTextEmbedding(
            model_name=model_name,
            cache_dir=cache_dir,
            threads=1,
            **kwargs,
        )



from typing import Iterable, Optional, Union, Any


from fastembed.common import ImageInput
from fastembed.common.model_description import DenseModelDescription
from fastembed.common.model_management import ModelManagement
from fastembed.common.types import NumpyArray


class LateInteractionMultimodalEmbeddingBase(ModelManagement[DenseModelDescription]):
    def __init__(
        self,
        model_name: str,
        cache_dir: Optional[str] = None,
        threads: Optional[int] = None,
        **kwargs: Any,
    ):
        self.model_name = model_name
        self.cache_dir = cache_dir
        self.threads = threads
        self._local_files_only = kwargs.pop("local_files_only", False)
        self._embedding_size: Optional[int] = None

    def embed_text(
        self,
        documents: Union[str, Iterable[str]],
        batch_size: int = 256,
        parallel: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterable[NumpyArray]:
        """
        Embeds a list of documents into a list of embeddings.

        Args:
            documents (Iterable[str]): The list of texts to embed.
            batch_size: Batch size for encoding -- higher values will use more memory, but be faster
            parallel:
                If > 1, data-parallel encoding will be used, recommended for offline encoding of large datasets.
                If 0, use all available cores.
                If None, don't use data-parallel processing, use default onnxruntime threading instead.
            **kwargs: Additional keyword argument to pass to the embed method.

        Yields:
            Iterable[NumpyArray]: The embeddings.
        """
        raise NotImplementedError()

    def embed_image(
        self,
        images: Union[ImageInput, Iterable[ImageInput]],
        batch_size: int = 16,
        parallel: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterable[NumpyArray]:
        """
        Encode a list of images into list of embeddings.
        Args:
            images: Iterator of image paths or single image path to embed
            batch_size: Batch size for encoding -- higher values will use more memory, but be faster
            parallel:
                If > 1, data-parallel encoding will be used, recommended for offline encoding of large datasets.
                If 0, use all available cores.
                If None, don't use data-parallel processing, use default onnxruntime threading instead.

        Returns:
            List of embeddings, one per image
        """
        raise NotImplementedError()

    @classmethod
    def get_embedding_size(cls, model_name: str) -> int:
        """Returns embedding size of the chosen model."""
        raise NotImplementedError("Subclasses must implement this method")

    @property
    def embedding_size(self) -> int:
        """Returns embedding size for the current model"""
        raise NotImplementedError("Subclasses must implement this method")



from collections import defaultdict
from typing import Iterable

from py_rust_stemmers import SnowballStemmer
import numpy as np
from tokenizers import Tokenizer
from numpy.typing import NDArray

from fastembed.common.types import NumpyArray


class VocabTokenizerBase:
    def tokenize(self, sentence: str) -> NumpyArray:
        raise NotImplementedError()

    def convert_ids_to_tokens(self, token_ids: NumpyArray) -> list[str]:
        raise NotImplementedError()


class VocabTokenizer(VocabTokenizerBase):
    def __init__(self, tokenizer: Tokenizer):
        self.tokenizer = tokenizer

    def tokenize(self, sentence: str) -> NumpyArray:
        return np.array(self.tokenizer.encode(sentence).ids)

    def convert_ids_to_tokens(self, token_ids: NumpyArray) -> list[str]:
        return [self.tokenizer.id_to_token(token_id) for token_id in token_ids]


class VocabResolver:
    def __init__(self, tokenizer: VocabTokenizerBase, stopwords: set[str], stemmer: SnowballStemmer):
        # Word to id mapping
        self.vocab: dict[str, int] = {}
        # Id to word mapping
        self.words: list[str] = []
        # Lemma to word mapping
        self.stem_mapping: dict[str, str] = {}
        self.tokenizer: VocabTokenizerBase = tokenizer
        self.stemmer = stemmer
        self.stopwords: set[str] = stopwords

    def tokenize(self, sentence: str) -> NumpyArray:
        return self.tokenizer.tokenize(sentence)

    def lookup_word(self, word_id: int) -> str:
        if word_id == 0:
            return "UNK"
        return self.words[word_id - 1]

    def convert_ids_to_tokens(self, token_ids: NumpyArray) -> list[str]:
        return self.tokenizer.convert_ids_to_tokens(token_ids)

    def vocab_size(self) -> int:
        # We need +1 for UNK token
        return len(self.vocab) + 1

    def save_vocab(self, path: str) -> None:
        with open(path, "w") as f:
            for word in self.words:
                f.write(word + "\n")

    def save_json_vocab(self, path: str) -> None:
        import json

        with open(path, "w") as f:
            json.dump({"vocab": self.words, "stem_mapping": self.stem_mapping}, f, indent=2)

    def load_json_vocab(self, path: str) -> None:
        import json

        with open(path, "r") as f:
            data = json.load(f)
            self.words = data["vocab"]
            self.vocab = {word: idx + 1 for idx, word in enumerate(self.words)}
            self.stem_mapping = data["stem_mapping"]

    def add_word(self, word: str) -> None:
        if word not in self.vocab:
            self.vocab[word] = len(self.vocab) + 1
            self.words.append(word)
            stem = self.stemmer.stem_word(word)
            if stem not in self.stem_mapping:
                self.stem_mapping[stem] = word
            else:
                existing_word = self.stem_mapping[stem]
                if len(existing_word) > len(word):
                    # Prefer shorter words for the same stem
                    # Example: "swim" is preferred over "swimming"
                    self.stem_mapping[stem] = word

    def load_vocab(self, path: str) -> None:
        with open(path, "r") as f:
            for line in f:
                self.add_word(line.strip())

    @classmethod
    def _reconstruct_bpe(
        cls, bpe_tokens: Iterable[tuple[int, str]]
    ) -> list[tuple[str, list[int]]]:
        result: list[tuple[str, list[int]]] = []
        acc: str = ""
        acc_idx: list[int] = []

        continuing_subword_prefix = "##"
        continuing_subword_prefix_len = len(continuing_subword_prefix)

        for idx, token in bpe_tokens:
            if token.startswith(continuing_subword_prefix):
                acc += token[continuing_subword_prefix_len:]
                acc_idx.append(idx)
            else:
                if acc:
                    result.append((acc, acc_idx))
                    acc_idx = []
                acc = token
                acc_idx.append(idx)

        if acc:
            result.append((acc, acc_idx))
        return result

    def resolve_tokens(
        self, token_ids: NDArray[np.int64]
    ) -> tuple[NDArray[np.int64], dict[int, int], dict[str, int], dict[str, list[str]]]:
        """
        Mark known tokens (including composed tokens) with vocab ids.

        Args:
            token_ids: (seq_len) - list of ids of tokens
                Example:
                    [
                        101,  3897, 19332, 12718, 23348,
                        1010,  1996,  7151,  2296, 4845,
                        2359,  2005,  4234,  1010,  4332,
                        2871,  3191,  2062, 102
                    ]

            returns:
                - token_ids with vocab ids
                    [
                        0,  151, 151, 0, 0,
                        912,  0,  0,  0, 332,
                        332,  332,  0,  7121,  191,
                        0,  0,  332, 0
                    ]
                - counts of each token
                    {
                        151: 1,
                        332: 3,
                        7121: 1,
                        191: 1,
                        912: 1
                    }
                - oov counts of each token
                    {
                        "the": 1,
                        "a": 1,
                        "[CLS]": 1,
                        "[SEP]": 1,
                        ...
                    }
                - forms of each token
                    {
                        "hello": ["hello"],
                        "world": ["worlds", "world", "worlding"],
                    }

        """
        tokens = self.convert_ids_to_tokens(token_ids)
        tokens_mapping = self._reconstruct_bpe(enumerate(tokens))

        counts: dict[int, int] = defaultdict(int)
        oov_count: dict[str, int] = defaultdict(int)

        forms: dict[str, list[str]] = defaultdict(list)

        for token, mapped_token_ids in tokens_mapping:
            vocab_id = 0
            if token in self.stopwords:
                vocab_id = 0
            elif token in self.vocab:
                vocab_id = self.vocab[token]
                forms[token].append(token)
            elif token in self.stem_mapping:
                vocab_id = self.vocab[self.stem_mapping[token]]
                forms[self.stem_mapping[token]].append(token)
            else:
                stem = self.stemmer.stem_word(token)
                if stem in self.stem_mapping:
                    vocab_id = self.vocab[self.stem_mapping[stem]]
                    forms[self.stem_mapping[stem]].append(token)

            for token_id in mapped_token_ids:
                token_ids[token_id] = vocab_id

            if vocab_id == 0:
                oov_count[token] += 1
            else:
                counts[vocab_id] += 1
        return token_ids, counts, oov_count, forms




from typing import Any, Iterable, Optional, Sequence, Type, Union

import numpy as np
from fastembed.common import OnnxProvider
from fastembed.common.onnx_model import OnnxOutputContext
from fastembed.common.utils import define_cache_dir
from fastembed.sparse.sparse_embedding_base import (
    SparseEmbedding,
    SparseTextEmbeddingBase,
)
from fastembed.text.onnx_text_model import OnnxTextModel, TextEmbeddingWorker
from fastembed.common.model_description import SparseModelDescription, ModelSource

supported_splade_models: list[SparseModelDescription] = [
    SparseModelDescription(
        model="prithivida/Splade_PP_en_v1",
        vocab_size=30522,
        description="Independent Implementation of SPLADE++ Model for English.",
        license="apache-2.0",
        size_in_GB=0.532,
        sources=ModelSource(hf="Qdrant/Splade_PP_en_v1"),
        model_file="model.onnx",
    ),
    SparseModelDescription(
        model="prithvida/Splade_PP_en_v1",
        vocab_size=30522,
        description="Independent Implementation of SPLADE++ Model for English.",
        license="apache-2.0",
        size_in_GB=0.532,
        sources=ModelSource(hf="Qdrant/Splade_PP_en_v1"),
        model_file="model.onnx",
    ),
]


class SpladePP(SparseTextEmbeddingBase, OnnxTextModel[SparseEmbedding]):
    def _post_process_onnx_output(
        self, output: OnnxOutputContext, **kwargs: Any
    ) -> Iterable[SparseEmbedding]:
        if output.attention_mask is None:
            raise ValueError("attention_mask must be provided for document post-processing")

        relu_log = np.log(1 + np.maximum(output.model_output, 0))

        weighted_log = relu_log * np.expand_dims(output.attention_mask, axis=-1)

        scores = np.max(weighted_log, axis=1)

        # Score matrix of shape (batch_size, vocab_size)
        # Most of the values are 0, only a few are non-zero
        for row_scores in scores:
            indices = row_scores.nonzero()[0]
            scores = row_scores[indices]
            yield SparseEmbedding(values=scores, indices=indices)

    @classmethod
    def _list_supported_models(cls) -> list[SparseModelDescription]:
        """Lists the supported models.

        Returns:
            list[SparseModelDescription]: A list of SparseModelDescription objects containing the model information.
        """
        return supported_splade_models

    def __init__(
        self,
        model_name: str,
        cache_dir: Optional[str] = None,
        threads: Optional[int] = None,
        providers: Optional[Sequence[OnnxProvider]] = None,
        cuda: bool = False,
        device_ids: Optional[list[int]] = None,
        lazy_load: bool = False,
        device_id: Optional[int] = None,
        specific_model_path: Optional[str] = None,
        **kwargs: Any,
    ):
        """
        Args:
            model_name (str): The name of the model to use.
            cache_dir (str, optional): The path to the cache directory.
                                       Can be set using the `FASTEMBED_CACHE_PATH` env variable.
                                       Defaults to `fastembed_cache` in the system's temp directory.
            threads (int, optional): The number of threads single onnxruntime session can use. Defaults to None.
            providers (Optional[Sequence[OnnxProvider]], optional): The list of onnxruntime providers to use.
                Mutually exclusive with the `cuda` and `device_ids` arguments. Defaults to None.
            cuda (bool, optional): Whether to use cuda for inference. Mutually exclusive with `providers`
                Defaults to False.
            device_ids (Optional[list[int]], optional): The list of device ids to use for data parallel processing in
                workers. Should be used with `cuda=True`, mutually exclusive with `providers`. Defaults to None.
            lazy_load (bool, optional): Whether to load the model during class initialization or on demand.
                Should be set to True when using multiple-gpu and parallel encoding. Defaults to False.
            device_id (Optional[int], optional): The device id to use for loading the model in the worker process.
            specific_model_path (Optional[str], optional): The specific path to the onnx model dir if it should be imported from somewhere else

        Raises:
            ValueError: If the model_name is not in the format <org>/<model> e.g. BAAI/bge-base-en.
        """
        super().__init__(model_name, cache_dir, threads, **kwargs)
        self.providers = providers
        self.lazy_load = lazy_load

        # List of device ids, that can be used for data parallel processing in workers
        self.device_ids = device_ids
        self.cuda = cuda

        # This device_id will be used if we need to load model in current process
        self.device_id: Optional[int] = None
        if device_id is not None:
            self.device_id = device_id
        elif self.device_ids is not None:
            self.device_id = self.device_ids[0]

        self.model_description = self._get_model_description(model_name)
        self.cache_dir = str(define_cache_dir(cache_dir))

        self._specific_model_path = specific_model_path
        self._model_dir = self.download_model(
            self.model_description,
            self.cache_dir,
            local_files_only=self._local_files_only,
            specific_model_path=self._specific_model_path,
        )

        if not self.lazy_load:
            self.load_onnx_model()

    def load_onnx_model(self) -> None:
        self._load_onnx_model(
            model_dir=self._model_dir,
            model_file=self.model_description.model_file,
            threads=self.threads,
            providers=self.providers,
            cuda=self.cuda,
            device_id=self.device_id,
        )

    def embed(
        self,
        documents: Union[str, Iterable[str]],
        batch_size: int = 256,
        parallel: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterable[SparseEmbedding]:
        """
        Encode a list of documents into list of embeddings.
        We use mean pooling with attention so that the model can handle variable-length inputs.

        Args:
            documents: Iterator of documents or single document to embed
            batch_size: Batch size for encoding -- higher values will use more memory, but be faster
            parallel:
                If > 1, data-parallel encoding will be used, recommended for offline encoding of large datasets.
                If 0, use all available cores.
                If None, don't use data-parallel processing, use default onnxruntime threading instead.

        Returns:
            List of embeddings, one per document
        """
        yield from self._embed_documents(
            model_name=self.model_name,
            cache_dir=str(self.cache_dir),
            documents=documents,
            batch_size=batch_size,
            parallel=parallel,
            providers=self.providers,
            cuda=self.cuda,
            device_ids=self.device_ids,
            local_files_only=self._local_files_only,
            specific_model_path=self._specific_model_path,
            **kwargs,
        )

    @classmethod
    def _get_worker_class(cls) -> Type[TextEmbeddingWorker[SparseEmbedding]]:
        return SpladePPEmbeddingWorker


class SpladePPEmbeddingWorker(TextEmbeddingWorker[SparseEmbedding]):
    def init_embedding(self, model_name: str, cache_dir: str, **kwargs: Any) -> SpladePP:
        return SpladePP(
            model_name=model_name,
            cache_dir=cache_dir,
            threads=1,
            **kwargs,
        )



from typing import Union

import numpy as np

from fastembed.common.types import NumpyArray
from fastembed.late_interaction.late_interaction_embedding_base import (
    LateInteractionTextEmbeddingBase,
)
from fastembed.late_interaction_multimodal.late_interaction_multimodal_embedding_base import (
    LateInteractionMultimodalEmbeddingBase,
)


MultiVectorModel = Union[LateInteractionTextEmbeddingBase, LateInteractionMultimodalEmbeddingBase]
MAX_HAMMING_DISTANCE = 65  # 64 bits + 1
POPCOUNT_LUT = np.array([bin(x).count("1") for x in range(256)], dtype=np.uint8)


def hamming_distance_matrix(ids: np.ndarray) -> np.ndarray:
    """Compute full Hamming distance matrix

    Args:
    ids: shape (n,) - array of ids, only size of the array matters

    Return:
        np.ndarray (n, n) - hamming distance matrix
    """
    n = len(ids)
    xor_vals = np.bitwise_xor(ids[:, None], ids[None, :])  # (n, n) uint64
    bytes_view = xor_vals.view(np.uint8).reshape(n, n, 8)  # (n, n, 8)
    return POPCOUNT_LUT[bytes_view].sum(axis=2)


class SimHashProjection:
    """
    SimHash projection component for MUVERA clustering.

    This class implements locality-sensitive hashing using random hyperplanes
    to partition the vector space into 2^k_sim clusters. Each vector is assigned
    to a cluster based on which side of k_sim random hyperplanes it falls on.

    Attributes:
        k_sim (int): Number of SimHash functions (hyperplanes)
        dim (int): Dimensionality of input vectors
        simhash_vectors (np.ndarray): Random hyperplane normal vectors of shape (dim, k_sim)
    """

    def __init__(self, k_sim: int, dim: int, random_generator: np.random.Generator):
        """
        Initialize SimHash projection with random hyperplanes.

        Args:
            k_sim (int): Number of SimHash functions, determines 2^k_sim clusters
            dim (int): Dimensionality of input vectors
            random_generator (np.random.Generator): Random number generator for reproducibility
        """
        self.k_sim = k_sim
        self.dim = dim
        # Generate k_sim random hyperplanes (normal vectors) from standard normal distribution
        self.simhash_vectors = random_generator.normal(size=(dim, k_sim))

    def get_cluster_ids(self, vectors: np.ndarray) -> np.ndarray:
        """
        Compute the cluster IDs for a given vector using SimHash.

        The cluster ID is determined by computing the dot product of the vector
        with each hyperplane normal vector, taking the sign, and interpreting
        the resulting binary string as an integer.

        Args:
            vectors (np.ndarray): Input vectors of shape (n, dim,)

        Returns:
            np.ndarray: Cluster IDs in range [0, 2^k_sim - 1]

        Raises:
            AssertionError: If a vector shape doesn't match expected dimensionality
        """
        dot_product = (
            vectors @ self.simhash_vectors
        )  # (token_num, dim) x (dim, k_sim) -> (token_num, k_sim)
        cluster_ids = (dot_product > 0) @ (1 << np.arange(self.k_sim))
        return cluster_ids


class Muvera:
    """
    MUVERA (Multi-Vector Retrieval Architecture) algorithm implementation.

    This class creates Fixed Dimensional Encodings (FDEs) from variable-length
    sequences of vectors by using SimHash clustering and random projections.
    The process involves:
    1. Clustering vectors using multiple SimHash projections
    2. Computing cluster centers (with different strategies for docs vs queries)
    3. Applying random projections for dimensionality reduction
    4. Concatenating results from all projections

    Attributes:
        k_sim (int): Number of SimHash functions per projection
        dim (int): Input vector dimensionality
        dim_proj (int): Output dimensionality after random projection
        r_reps (int): Number of random projection repetitions
        random_seed (int): Random seed for consistent random matrix generation
        simhash_projections (List[SimHashProjection]): SimHash instances for clustering
        dim_reduction_projections (np.ndarray): Random projection matrices of shape (R_reps, d, d_proj)
    """

    def __init__(
        self,
        dim: int,
        k_sim: int = 5,
        dim_proj: int = 16,
        r_reps: int = 20,
        random_seed: int = 42,
    ):
        """
        Initialize MUVERA algorithm with specified parameters.

        Args:
            dim (int): Dimensionality of individual input vectors
            k_sim (int, optional): Number of SimHash functions (creates 2^k_sim clusters).
                                   Defaults to 5.
            dim_proj (int, optional): Dimensionality after random projection (must be <= dim).
                                    Defaults to 16.
            r_reps (int, optional): Number of random projection repetitions for robustness.
                                    Defaults to 20.
            random_seed (int, optional): Seed for random number generator to ensure
                                         reproducible results. Defaults to 42.

        Raises:
            ValueError: If dim_proj > dim (cannot project to higher dimensionality)
        """
        if dim_proj > dim:
            raise ValueError(
                f"Cannot project to a higher dimensionality (dim_proj={dim_proj} > dim={dim})"
            )

        self.k_sim = k_sim
        self.dim = dim
        self.dim_proj = dim_proj
        self.r_reps = r_reps
        # Create r_reps independent SimHash projections for robustness
        generator = np.random.default_rng(random_seed)
        self.simhash_projections = [
            SimHashProjection(k_sim=self.k_sim, dim=self.dim, random_generator=generator)
            for _ in range(r_reps)
        ]
        # Random projection matrices with entries from {-1, +1} for each repetition
        self.dim_reduction_projections = generator.choice([-1, 1], size=(r_reps, dim, dim_proj))

    @classmethod
    def from_multivector_model(
        cls,
        model: MultiVectorModel,
        k_sim: int = 5,
        dim_proj: int = 16,
        r_reps: int = 20,  # noqa[naming]
        random_seed: int = 42,
    ) -> "Muvera":
        """
        Create a Muvera instance from a multi-vector embedding model.

        This class method provides a convenient way to initialize a MUVERA
        that is compatible with a given multi-vector model by automatically extracting
        the embedding dimensionality from the model.

        Args:
            model (MultiVectorModel): A late interaction text or multimodal embedding model
                                    that provides multi-vector embeddings. Must have an
                                    `embedding_size` attribute specifying the dimensionality
                                    of individual vectors.
            k_sim (int, optional): Number of SimHash functions (creates 2^k_sim clusters).
                                   Defaults to 5.
            dim_proj (int, optional): Dimensionality after random projection (must be <= model's
                                    embedding_size). Defaults to 16.
            r_reps (int, optional): Number of random projection repetitions for robustness.
                                    Defaults to 20.
            random_seed (int, optional): Seed for random number generator to ensure
                                         reproducible results. Defaults to 42.

        Returns:
            Muvera: A configured MUVERA instance ready to process embeddings from the given model.

        Raises:
            ValueError: If dim_proj > model.embedding_size (cannot project to higher dimensionality)

        Example:
            >>> from fastembed import LateInteractionTextEmbedding
            >>> model = LateInteractionTextEmbedding(model_name="colbert-ir/colbertv2.0")
            >>> muvera = Muvera.from_multivector_model(
            ...     model=model,
            ...     k_sim=6,
            ...     dim_proj=32
            ... )
            >>> # Now use postprocessor with embeddings from the model
            >>> embeddings = np.array(list(model.embed(["sample text"])))
            >>> fde = muvera.process_document(embeddings[0])
        """
        return cls(
            dim=model.embedding_size,
            k_sim=k_sim,
            dim_proj=dim_proj,
            r_reps=r_reps,
            random_seed=random_seed,
        )

    def _get_output_dimension(self) -> int:
        """
        Get the output dimension of the MUVERA algorithm.

        Returns:
            int: Output dimension (r_reps * num_partitions * dim_proj) where b = 2^k_sim
        """
        num_partitions = 2**self.k_sim
        return self.r_reps * num_partitions * self.dim_proj

    @property
    def embedding_size(self) -> int:
        return self._get_output_dimension()

    def process_document(self, vectors: NumpyArray) -> NumpyArray:
        """
        Encode a document's vectors into a Fixed Dimensional Encoding (FDE).

        Uses document-specific settings: normalizes cluster centers by vector count
        and fills empty clusters using Hamming distance-based selection.

        Args:
            vectors (NumpyArray): Document vectors of shape (n_tokens, dim)

        Returns:
            NumpyArray: Fixed dimensional encodings of shape (r_reps * b * dim_proj,)
        """
        return self.process(vectors, fill_empty_clusters=True, normalize_by_count=True)

    def process_query(self, vectors: NumpyArray) -> NumpyArray:
        """
        Encode a query's vectors into a Fixed Dimensional Encoding (FDE).

        Uses query-specific settings: no normalization by count and no empty
        cluster filling to preserve query vector magnitudes.

        Args:
            vectors (NumpyArray]): Query vectors of shape (n_tokens, dim)

        Returns:
            NumpyArray: Fixed dimensional encoding of shape (r_reps * b * dim_proj,)
        """
        return self.process(vectors, fill_empty_clusters=False, normalize_by_count=False)

    def process(
        self,
        vectors: NumpyArray,
        fill_empty_clusters: bool = True,
        normalize_by_count: bool = True,
    ) -> NumpyArray:
        """
        Core encoding method that transforms variable-length vector sequences into FDEs.

        The encoding process:
        1. For each of r_reps random projections:
           a. Assign vectors to clusters using SimHash
           b. Compute cluster centers (sum of vectors in each cluster)
           c. Optionally normalize by cluster size
           d. Fill empty clusters using Hamming distance if requested
           e. Apply random projection for dimensionality reduction
           f. Flatten cluster centers into a vector
        2. Concatenate all projection results

        Args:
            vectors (np.ndarray): Input vectors of shape (n_vectors, dim)
            fill_empty_clusters (bool): Whether to fill empty clusters using nearest
                                      vectors based on Hamming distance of cluster IDs
            normalize_by_count (bool): Whether to normalize cluster centers by the
                                     number of vectors assigned to each cluster

        Returns:
            np.ndarray: Fixed dimensional encoding of shape (r_reps * b * dim_proj)
                        where B = 2^k_sim is the number of clusters

        Raises:
            AssertionError: If input vectors don't have expected dimensionality
        """
        assert (
            vectors.shape[1] == self.dim
        ), f"Expected vectors of shape (n, {self.dim}), got {vectors.shape}"

        # Store results from each random projection
        output_vectors = []

        # num of space partitions in SimHash
        num_partitions = 2**self.k_sim
        cluster_center_ids = np.arange(num_partitions)
        precomputed_hamming_matrix = (
            hamming_distance_matrix(cluster_center_ids) if fill_empty_clusters else None
        )

        for projection_index, simhash in enumerate(self.simhash_projections):
            # Initialize cluster centers and count vectors assigned to each cluster
            cluster_centers = np.zeros((num_partitions, self.dim))
            cluster_center_id_to_vectors: dict[int, list[int]] = {
                cluster_center_id: [] for cluster_center_id in cluster_center_ids
            }
            cluster_vector_counts = None
            empty_mask = None

            # Assign each vector to its cluster and accumulate cluster centers
            vector_cluster_ids = simhash.get_cluster_ids(vectors)
            for cluster_id, (vec_idx, vec) in zip(vector_cluster_ids, enumerate(vectors)):
                cluster_centers[cluster_id] += vec
                cluster_center_id_to_vectors[cluster_id].append(vec_idx)

            if normalize_by_count or fill_empty_clusters:
                cluster_vector_counts = np.bincount(vector_cluster_ids, minlength=num_partitions)
                empty_mask = cluster_vector_counts == 0

            if normalize_by_count:
                assert empty_mask is not None
                assert cluster_vector_counts is not None
                non_empty_mask = ~empty_mask
                cluster_centers[non_empty_mask] /= cluster_vector_counts[non_empty_mask][:, None]

            # Fill empty clusters using vectors with minimum Hamming distance
            if fill_empty_clusters:
                assert empty_mask is not None
                assert precomputed_hamming_matrix is not None
                masked_hamming = np.where(
                    empty_mask[None, :], MAX_HAMMING_DISTANCE, precomputed_hamming_matrix
                )
                nearest_non_empty = np.argmin(masked_hamming, axis=1)
                fill_vectors = np.array(
                    [
                        vectors[cluster_center_id_to_vectors[cluster_id][0]]
                        for cluster_id in nearest_non_empty[empty_mask]
                    ]
                ).reshape(-1, self.dim)
                cluster_centers[empty_mask] = fill_vectors

            # Apply random projection for dimensionality reduction if needed
            if self.dim_proj < self.dim:
                dim_reduction_projection = self.dim_reduction_projections[
                    projection_index
                ]  # Get projection matrix for this repetition
                projected_centers = (1 / np.sqrt(self.dim_proj)) * (
                    cluster_centers @ dim_reduction_projection
                )

                # Flatten cluster centers into a single vector and add to output
                output_vectors.append(projected_centers.flatten())
                continue

            # If no projection needed (dim_proj == dim), use original cluster centers
            output_vectors.append(cluster_centers.flatten())

        # Concatenate results from all R_reps projections into final FDE
        return np.concatenate(output_vectors)


if __name__ == "__main__":
    v_arrs = np.random.randn(10, 100, 128)
    muvera = Muvera(128, 4, 8, 20, 42)

    for v_arr in v_arrs:
        muvera.process(v_arr)  # type: ignore



from typing import Any, Iterable, Type

from fastembed.common.types import NumpyArray
from fastembed.common.onnx_model import OnnxOutputContext
from fastembed.text.onnx_embedding import OnnxTextEmbedding, OnnxTextEmbeddingWorker
from fastembed.common.model_description import DenseModelDescription, ModelSource

supported_clip_models: list[DenseModelDescription] = [
    DenseModelDescription(
        model="Qdrant/clip-ViT-B-32-text",
        dim=512,
        description=(
            "Text embeddings, Multimodal (text&image), English, 77 input tokens truncation, "
            "Prefixes for queries/documents: not necessary, 2021 year"
        ),
        license="mit",
        size_in_GB=0.25,
        sources=ModelSource(hf="Qdrant/clip-ViT-B-32-text"),
        model_file="model.onnx",
    ),
]


class CLIPOnnxEmbedding(OnnxTextEmbedding):
    @classmethod
    def _get_worker_class(cls) -> Type[OnnxTextEmbeddingWorker]:
        return CLIPEmbeddingWorker

    @classmethod
    def _list_supported_models(cls) -> list[DenseModelDescription]:
        """Lists the supported models.

        Returns:
            list[DenseModelDescription]: A list of DenseModelDescription objects containing the model information.
        """
        return supported_clip_models

    def _post_process_onnx_output(
        self, output: OnnxOutputContext, **kwargs: Any
    ) -> Iterable[NumpyArray]:
        return output.model_output


class CLIPEmbeddingWorker(OnnxTextEmbeddingWorker):
    def init_embedding(
        self,
        model_name: str,
        cache_dir: str,
        **kwargs: Any,
    ) -> OnnxTextEmbedding:
        return CLIPOnnxEmbedding(
            model_name=model_name,
            cache_dir=cache_dir,
            threads=1,
            **kwargs,
        )



import importlib.metadata

from fastembed.image import ImageEmbedding
from fastembed.late_interaction import LateInteractionTextEmbedding
from fastembed.late_interaction_multimodal import LateInteractionMultimodalEmbedding
from fastembed.sparse import SparseEmbedding, SparseTextEmbedding
from fastembed.text import TextEmbedding

try:
    version = importlib.metadata.version("fastembed")
except importlib.metadata.PackageNotFoundError as _:
    version = importlib.metadata.version("fastembed-gpu")

__version__ = version
__all__ = [
    "TextEmbedding",
    "SparseTextEmbedding",
    "SparseEmbedding",
    "ImageEmbedding",
    "LateInteractionTextEmbedding",
    "LateInteractionMultimodalEmbedding",
]



from typing import Optional, Any

from loguru import logger

from fastembed import TextEmbedding

logger.warning(
    "DefaultEmbedding, FlagEmbedding, JinaEmbedding are deprecated."
    "Use from fastembed import TextEmbedding instead."
)

DefaultEmbedding = TextEmbedding
FlagEmbedding = TextEmbedding


class JinaEmbedding(TextEmbedding):
    def __init__(
        self,
        model_name: str = "jinaai/jina-embeddings-v2-base-en",
        cache_dir: Optional[str] = None,
        threads: Optional[int] = None,
        **kwargs: Any,
    ):
        super().__init__(model_name, cache_dir, threads, **kwargs)



import os
from multiprocessing import get_all_start_methods
from pathlib import Path
from typing import Any, Iterable, Optional, Sequence, Type, Union

import numpy as np
from numpy.typing import NDArray
from tokenizers import Encoding, Tokenizer

from fastembed.common.types import NumpyArray, OnnxProvider
from fastembed.common.onnx_model import EmbeddingWorker, OnnxModel, OnnxOutputContext, T
from fastembed.common.preprocessor_utils import load_tokenizer
from fastembed.common.utils import iter_batch
from fastembed.parallel_processor import ParallelWorkerPool


class OnnxTextModel(OnnxModel[T]):
    ONNX_OUTPUT_NAMES: Optional[list[str]] = None

    @classmethod
    def _get_worker_class(cls) -> Type["TextEmbeddingWorker[T]"]:
        raise NotImplementedError("Subclasses must implement this method")

    def _post_process_onnx_output(self, output: OnnxOutputContext, **kwargs: Any) -> Iterable[T]:
        """Post-process the ONNX model output to convert it into a usable format.

        Args:
            output (OnnxOutputContext): The raw output from the ONNX model.
            **kwargs: Additional keyword arguments that may be needed by specific implementations.

        Returns:
            Iterable[T]: Post-processed output as an iterable of type T.
        """
        raise NotImplementedError("Subclasses must implement this method")

    def __init__(self) -> None:
        super().__init__()
        self.tokenizer: Optional[Tokenizer] = None
        self.special_token_to_id: dict[str, int] = {}

    def _preprocess_onnx_input(
        self, onnx_input: dict[str, NumpyArray], **kwargs: Any
    ) -> dict[str, Union[NumpyArray, NDArray[np.int64]]]:
        """
        Preprocess the onnx input.
        """
        return onnx_input

    def _load_onnx_model(
        self,
        model_dir: Path,
        model_file: str,
        threads: Optional[int],
        providers: Optional[Sequence[OnnxProvider]] = None,
        cuda: bool = False,
        device_id: Optional[int] = None,
    ) -> None:
        super()._load_onnx_model(
            model_dir=model_dir,
            model_file=model_file,
            threads=threads,
            providers=providers,
            cuda=cuda,
            device_id=device_id,
        )
        self.tokenizer, self.special_token_to_id = load_tokenizer(model_dir=model_dir)

    def load_onnx_model(self) -> None:
        raise NotImplementedError("Subclasses must implement this method")

    def tokenize(self, documents: list[str], **kwargs: Any) -> list[Encoding]:
        return self.tokenizer.encode_batch(documents)  # type: ignore[union-attr]

    def onnx_embed(
        self,
        documents: list[str],
        **kwargs: Any,
    ) -> OnnxOutputContext:
        encoded = self.tokenize(documents, **kwargs)
        input_ids = np.array([e.ids for e in encoded])
        attention_mask = np.array([e.attention_mask for e in encoded])
        input_names = {node.name for node in self.model.get_inputs()}  # type: ignore[union-attr]
        onnx_input: dict[str, NumpyArray] = {
            "input_ids": np.array(input_ids, dtype=np.int64),
        }
        if "attention_mask" in input_names:
            onnx_input["attention_mask"] = np.array(attention_mask, dtype=np.int64)
        if "token_type_ids" in input_names:
            onnx_input["token_type_ids"] = np.array(
                [np.zeros(len(e), dtype=np.int64) for e in input_ids], dtype=np.int64
            )
        onnx_input = self._preprocess_onnx_input(onnx_input, **kwargs)

        model_output = self.model.run(self.ONNX_OUTPUT_NAMES, onnx_input)  # type: ignore[union-attr]
        return OnnxOutputContext(
            model_output=model_output[0],
            attention_mask=onnx_input.get("attention_mask", attention_mask),
            input_ids=onnx_input.get("input_ids", input_ids),
        )

    def _embed_documents(
        self,
        model_name: str,
        cache_dir: str,
        documents: Union[str, Iterable[str]],
        batch_size: int = 256,
        parallel: Optional[int] = None,
        providers: Optional[Sequence[OnnxProvider]] = None,
        cuda: bool = False,
        device_ids: Optional[list[int]] = None,
        local_files_only: bool = False,
        specific_model_path: Optional[str] = None,
        **kwargs: Any,
    ) -> Iterable[T]:
        is_small = False

        if isinstance(documents, str):
            documents = [documents]
            is_small = True

        if isinstance(documents, list):
            if len(documents) < batch_size:
                is_small = True

        if parallel is None or is_small:
            if not hasattr(self, "model") or self.model is None:
                self.load_onnx_model()
            for batch in iter_batch(documents, batch_size):
                yield from self._post_process_onnx_output(
                    self.onnx_embed(batch, **kwargs), **kwargs
                )
        else:
            if parallel == 0:
                parallel = os.cpu_count()

            start_method = "forkserver" if "forkserver" in get_all_start_methods() else "spawn"
            params = {
                "model_name": model_name,
                "cache_dir": cache_dir,
                "providers": providers,
                "local_files_only": local_files_only,
                "specific_model_path": specific_model_path,
                **kwargs,
            }

            pool = ParallelWorkerPool(
                num_workers=parallel or 1,
                worker=self._get_worker_class(),
                cuda=cuda,
                device_ids=device_ids,
                start_method=start_method,
            )
            for batch in pool.ordered_map(iter_batch(documents, batch_size), **params):
                yield from self._post_process_onnx_output(batch, **kwargs)  # type: ignore


class TextEmbeddingWorker(EmbeddingWorker[T]):
    def process(self, items: Iterable[tuple[int, Any]]) -> Iterable[tuple[int, OnnxOutputContext]]:
        for idx, batch in items:
            onnx_output = self.model.onnx_embed(batch)
            yield idx, onnx_output



from typing import Any, Type

from fastembed.common.types import NumpyArray
from fastembed.late_interaction.colbert import Colbert, ColbertEmbeddingWorker
from fastembed.common.model_description import DenseModelDescription, ModelSource

supported_jina_colbert_models: list[DenseModelDescription] = [
    DenseModelDescription(
        model="jinaai/jina-colbert-v2",
        dim=128,
        description="New model that expands capabilities of colbert-v1 with multilingual and context length of 8192, 2024 year",
        license="cc-by-nc-4.0",
        size_in_GB=2.24,
        sources=ModelSource(hf="jinaai/jina-colbert-v2"),
        model_file="onnx/model.onnx",
        additional_files=["onnx/model.onnx_data"],
    )
]


class JinaColbert(Colbert):
    QUERY_MARKER_TOKEN_ID = 250002
    DOCUMENT_MARKER_TOKEN_ID = 250003
    MIN_QUERY_LENGTH = 31  # it's 32, we add one additional special token in the beginning
    MASK_TOKEN = "<mask>"

    @classmethod
    def _get_worker_class(cls) -> Type[ColbertEmbeddingWorker]:
        return JinaColbertEmbeddingWorker

    @classmethod
    def _list_supported_models(cls) -> list[DenseModelDescription]:
        """Lists the supported models.

        Returns:
            list[DenseModelDescription]: A list of DenseModelDescription objects containing the model information.
        """
        return supported_jina_colbert_models

    def _preprocess_onnx_input(
        self, onnx_input: dict[str, NumpyArray], is_doc: bool = True, **kwargs: Any
    ) -> dict[str, NumpyArray]:
        onnx_input = super()._preprocess_onnx_input(onnx_input, is_doc)

        # the attention mask for jina-colbert-v2 is always 1 in queries
        if not is_doc:
            onnx_input["attention_mask"][:] = 1
        return onnx_input


class JinaColbertEmbeddingWorker(ColbertEmbeddingWorker):
    def init_embedding(self, model_name: str, cache_dir: str, **kwargs: Any) -> JinaColbert:
        return JinaColbert(
            model_name=model_name,
            cache_dir=cache_dir,
            threads=1,
            **kwargs,
        )



from dataclasses import asdict
from typing import Union, Iterable, Optional, Any, Type

from fastembed.common.model_description import DenseModelDescription, ModelSource
from fastembed.common.onnx_model import OnnxOutputContext
from fastembed.common.types import NumpyArray
from fastembed.late_interaction.late_interaction_embedding_base import (
    LateInteractionTextEmbeddingBase,
)
from fastembed.text.onnx_embedding import OnnxTextEmbedding
from fastembed.text.onnx_text_model import TextEmbeddingWorker


supported_token_embeddings_models = [
    DenseModelDescription(
        model="jinaai/jina-embeddings-v2-small-en-tokens",
        dim=512,
        description="Text embeddings, Unimodal (text), English, 8192 input tokens truncation,"
        " Prefixes for queries/documents: not necessary, 2023 year.",
        license="apache-2.0",
        size_in_GB=0.12,
        sources=ModelSource(hf="xenova/jina-embeddings-v2-small-en"),
        model_file="onnx/model.onnx",
    ),
]


class TokenEmbeddingsModel(OnnxTextEmbedding, LateInteractionTextEmbeddingBase):
    @classmethod
    def _list_supported_models(cls) -> list[DenseModelDescription]:
        """Lists the supported models.

        Returns:
            list[DenseModelDescription]: A list of DenseModelDescription objects containing the model information.
        """
        return supported_token_embeddings_models

    @classmethod
    def list_supported_models(cls) -> list[dict[str, Any]]:
        """Lists the supported models.

        Returns:
            list[dict[str, Any]]: A list of dictionaries containing the model information.
        """
        return [asdict(model) for model in cls._list_supported_models()]

    @classmethod
    def _get_worker_class(cls) -> Type[TextEmbeddingWorker[NumpyArray]]:
        return TokensEmbeddingWorker

    def _post_process_onnx_output(
        self, output: OnnxOutputContext, **kwargs: Any
    ) -> Iterable[NumpyArray]:
        # Size: (batch_size, sequence_length, hidden_size)
        embeddings = output.model_output
        # Size: (batch_size, sequence_length)
        assert output.attention_mask is not None
        masks = output.attention_mask

        # For each document we only select those embeddings that are not masked out
        for i in range(embeddings.shape[0]):
            yield embeddings[i, masks[i] == 1]

    def embed(
        self,
        documents: Union[str, Iterable[str]],
        batch_size: int = 256,
        parallel: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterable[NumpyArray]:
        yield from super().embed(documents, batch_size=batch_size, parallel=parallel, **kwargs)


class TokensEmbeddingWorker(TextEmbeddingWorker[NumpyArray]):
    def init_embedding(
        self, model_name: str, cache_dir: str, **kwargs: Any
    ) -> TokenEmbeddingsModel:
        return TokenEmbeddingsModel(
            model_name=model_name,
            cache_dir=cache_dir,
            threads=1,
            **kwargs,
        )



from typing import Dict, List, Set
from py_rust_stemmers import SnowballStemmer
from fastembed.common.utils import get_all_punctuation, remove_non_alphanumeric
import mmh3
import copy
from dataclasses import dataclass

import numpy as np

from fastembed.sparse.sparse_embedding_base import SparseEmbedding

GAP = 32000
INT32_MAX = 2**31 - 1


@dataclass
class WordEmbedding:
    word: str
    forms: List[str]
    count: int
    word_id: int
    embedding: List[float]


class SparseVectorConverter:
    def __init__(
        self,
        stopwords: Set[str],
        stemmer: SnowballStemmer,
        k: float = 1.2,
        b: float = 0.75,
        avg_len: float = 150.0,
    ):
        punctuation = set(get_all_punctuation())
        special_tokens = {"[CLS]", "[SEP]", "[PAD]", "[UNK]", "[MASK]"}

        self.stemmer = stemmer
        self.unwanted_tokens = punctuation | special_tokens | stopwords

        self.k = k
        self.b = b
        self.avg_len = avg_len

    @classmethod
    def unkn_word_token_id(
        cls, word: str, shift: int
    ) -> int:  # 2-3 words can collide in 1 index with this mapping, not considering mm3 collisions
        token_hash = abs(mmh3.hash(word))

        range_size = INT32_MAX - shift
        remapped_hash = shift + (token_hash % range_size)

        return remapped_hash

    def bm25_tf(self, num_occurrences: int, sentence_len: int) -> float:
        res = num_occurrences * (self.k + 1)
        res /= num_occurrences + self.k * (1 - self.b + self.b * sentence_len / self.avg_len)
        return res

    @classmethod
    def normalize_vector(cls, vector: List[float]) -> List[float]:
        norm = sum([x**2 for x in vector]) ** 0.5
        if norm < 1e-8:
            return vector
        return [x / norm for x in vector]

    def clean_words(
        self, sentence_embedding: Dict[str, WordEmbedding], token_max_length: int = 40
    ) -> Dict[str, WordEmbedding]:
        """
        Clean miniCOIL-produced sentence_embedding, as unknown to the miniCOIL's stemmer tokens should fully resemble
        our BM25 token representation.

        sentence_embedding = {"9°": {"word": "9°", "word_id": -1, "count": 2, "embedding": [1], "forms": ["9°"]},
                "9": {"word": "9", "word_id": -1, "count": 2, "embedding": [1], "forms": ["9"]},
                "bat": {"word": "bat", "word_id": 2, "count": 3, "embedding": [0.2, 0.1, -0.2, -0.2], "forms": ["bats", "bat"]},
                "9°9": {"word": "9°9", "word_id": -1, "count": 1, "embedding": [1], "forms": ["9°9"]},
                "screech": {"word": "screech", "word_id": -1, "count": 1, "embedding": [1], "forms": ["screech"]},
                "screeched": {"word": "screeched", "word_id": -1, "count": 1, "embedding": [1], "forms": ["screeched"]}
                }
        cleaned_embedding_ground_truth = {
                "9": {"word": "9", "word_id": -1, "count": 6, "embedding": [1], "forms": ["9°", "9", "9°9", "9°9"]},
                "bat": {"word": "bat", "word_id": 2, "count": 3, "embedding": [0.2, 0.1, -0.2, -0.2], "forms": ["bats", "bat"]},
                "screech": {"word": "screech", "word_id": -1, "count": 2, "embedding": [1], "forms": ["screech", "screeched"]}
                }
        """

        new_sentence_embedding: Dict[str, WordEmbedding] = {}

        for word, embedding in sentence_embedding.items():
            # embedding = {
            #     "word": "vector",
            #     "forms": ["vector", "vectors"],
            #     "count": 2,
            #     "word_id": 1231,
            #     "embedding": [0.1, 0.2, 0.3, 0.4]
            # }
            if embedding.word_id > 0:
                # Known word, no need to clean
                new_sentence_embedding[word] = embedding
            else:
                # Unknown word
                if word in self.unwanted_tokens:
                    continue

                # Example complex word split:
                # word = `word^vec`
                word_cleaned = remove_non_alphanumeric(word).strip()
                # word_cleaned = `word vec`

                if len(word_cleaned) > 0:
                    # Subwords: ['word', 'vec']
                    for subword in word_cleaned.split():
                        stemmed_subword: str = self.stemmer.stem_word(subword)
                        if (
                            len(stemmed_subword) <= token_max_length
                            and stemmed_subword not in self.unwanted_tokens
                        ):
                            if stemmed_subword not in new_sentence_embedding:
                                new_sentence_embedding[stemmed_subword] = copy.deepcopy(embedding)
                                new_sentence_embedding[stemmed_subword].word = stemmed_subword
                            else:
                                new_sentence_embedding[stemmed_subword].count += embedding.count
                                new_sentence_embedding[stemmed_subword].forms += embedding.forms

        return new_sentence_embedding

    def embedding_to_vector(
        self,
        sentence_embedding: Dict[str, WordEmbedding],
        embedding_size: int,
        vocab_size: int,
    ) -> SparseEmbedding:
        """
        Convert miniCOIL sentence embedding to Qdrant sparse vector

        Example input:

        ```
        {
            "vector": WordEmbedding({ // Vocabulary word, encoded with miniCOIL normally
                "word": "vector",
                "forms": ["vector", "vectors"],
                "count": 2,
                "word_id": 1231,
                "embedding": [0.1, 0.2, 0.3, 0.4]
            }),
            "axiotic": WordEmbedding({ // Out-of-vocabulary word, fallback to BM25
                "word": "axiotic",
                "forms": ["axiotics"],
                "count": 1,
                "word_id": -1,
            })
        }
        ```

        """

        indices: List[int] = []
        values: List[float] = []
        
        # Example:
        # vocab_size = 10000
        # embedding_size = 4
        # GAP = 32000
        # 
        # We want to start random words section from the bucket, that is guaranteed to not
        # include any vocab words.
        # We need (vocab_size * embedding_size) slots for vocab words.
        # Therefore we need (vocab_size * embedding_size) // GAP + 1 buckets for vocab words.
        # Therefore, we can start random words from bucket (vocab_size * embedding_size) // GAP + 1 + 1

        # ID at which the scope of OOV words starts
        unknown_words_shift = (
            (vocab_size * embedding_size) // GAP + 2
        ) * GAP
        sentence_embedding_cleaned = self.clean_words(sentence_embedding)

        # Calculate sentence length after cleaning
        sentence_len = 0
        for embedding in sentence_embedding_cleaned.values():
            sentence_len += embedding.count

        for embedding in sentence_embedding_cleaned.values():
            word_id = embedding.word_id
            num_occurrences = embedding.count
            tf = self.bm25_tf(num_occurrences, sentence_len)
            if (
                word_id > 0
            ):  # miniCOIL starts with ID 1, we generally won't have word_id == 0 (UNK), as we don't add
                # these words to sentence_embedding
                embedding_values = embedding.embedding
                normalized_embedding = self.normalize_vector(embedding_values)

                for val_id, value in enumerate(normalized_embedding):
                    indices.append(
                        word_id * embedding_size + val_id
                    )  # since miniCOIL IDs start with 1
                    values.append(value * tf)
            else:
                indices.append(self.unkn_word_token_id(embedding.word, unknown_words_shift))
                values.append(tf)

        return SparseEmbedding(
            indices=np.array(indices, dtype=np.int32),
            values=np.array(values, dtype=np.float32),
        )

    def embedding_to_vector_query(
        self,
        sentence_embedding: Dict[str, WordEmbedding],
        embedding_size: int,
        vocab_size: int,
    ) -> SparseEmbedding:
        """
        Same as `embedding_to_vector`, but no TF
        """

        indices: List[int] = []
        values: List[float] = []

        # ID at which the scope of OOV words starts
        unknown_words_shift = ((vocab_size * embedding_size) // GAP + 2) * GAP

        sentence_embedding_cleaned = self.clean_words(sentence_embedding)

        for embedding in sentence_embedding_cleaned.values():
            word_id = embedding.word_id
            tf = 1.0

            if word_id >= 0:  # miniCOIL starts with ID 1
                embedding_values = embedding.embedding
                normalized_embedding = self.normalize_vector(embedding_values)

                for val_id, value in enumerate(normalized_embedding):
                    indices.append(
                        word_id * embedding_size + val_id
                    )  # since miniCOIL IDs start with 1
                    values.append(value * tf)
            else:
                indices.append(self.unkn_word_token_id(embedding.word, unknown_words_shift))
                values.append(tf)

        return SparseEmbedding(
            indices=np.array(indices, dtype=np.int32),
            values=np.array(values, dtype=np.float32),
        )



"""
Pure numpy implementation of encoder model for a single word.

This model is not trainable, and should only be used for inference.
"""

import numpy as np
from fastembed.common.types import NumpyArray


class Encoder:
    """
    Encoder(768, 4, 10000)

    Will look like this:


                                         Per-word
                                         Encoder Matrix
     ┌─────────────────────┐
     │ Token Embedding(768)├──────┐      (10k, 768, 4)
     └─────────────────────┘      │         ┌─────────┐
                                  │         │         │
     ┌─────────────────────┐      │       ┌─┴───────┐ │
     │                     │      │       │         │ │
     └─────────────────────┘      │     ┌─┴───────┐ │ │      ┌─────────┐
                                  └────►│         │ │ ├─────►│Tanh     │
     ┌─────────────────────┐            │         │ │ │      └─────────┘
     │                     │            │         │ ├─┘
     └─────────────────────┘            │         ├─┘
                                        │         │
     ┌─────────────────────┐            └─────────┘
     │                     │
     └─────────────────────┘

     Final linear transformation is accompanied by a non-linear activation function: Tanh.

     Tanh is used to ensure that the output is in the range [-1, 1].
     It would be easier to visually interpret the output of the model, assuming that each dimension
     would need to encode a type of semantic cluster.
    """

    def __init__(
        self,
        weights: NumpyArray,
    ):
        self.weights = weights
        self.vocab_size, self.input_dim, self.output_dim = weights.shape

        self.encoder_weights: NumpyArray = weights

        # Activation function
        self.activation = np.tanh

    @staticmethod
    def convert_vocab_ids(vocab_ids: NumpyArray) -> NumpyArray:
        """
        Convert vocab_ids of shape (batch_size, seq_len) into (batch_size, seq_len, 2)
        by appending batch_id alongside each vocab_id.
        """
        batch_size, seq_len = vocab_ids.shape
        batch_ids = np.arange(batch_size, dtype=vocab_ids.dtype).reshape(batch_size, 1)
        batch_ids = np.repeat(batch_ids, seq_len, axis=1)
        # Stack vocab_ids and batch_ids along the last dimension
        combined: NumpyArray = np.stack((vocab_ids, batch_ids), axis=2).astype(np.int32)
        return combined

    @classmethod
    def avg_by_vocab_ids(
        cls, vocab_ids: NumpyArray, embeddings: NumpyArray
    ) -> tuple[NumpyArray, NumpyArray]:
        """
        Takes:
            vocab_ids: (batch_size, seq_len) int array
            embeddings: (batch_size, seq_len, input_dim) float array

        Returns:
            unique_flattened_vocab_ids: (total_unique, 2) array of [vocab_id, batch_id]
            unique_flattened_embeddings: (total_unique, input_dim) averaged embeddings
        """
        input_dim = embeddings.shape[2]

        # Flatten vocab_ids and embeddings
        # flattened_vocab_ids: (batch_size*seq_len, 2)
        flattened_vocab_ids = cls.convert_vocab_ids(vocab_ids).reshape(-1, 2)

        # flattened_embeddings: (batch_size*seq_len, input_dim)
        flattened_embeddings = embeddings.reshape(-1, input_dim)

        # Find unique (vocab_id, batch_id) pairs
        unique_flattened_vocab_ids, inverse_indices = np.unique(
            flattened_vocab_ids, axis=0, return_inverse=True
        )

        # Prepare arrays to accumulate sums
        unique_count = unique_flattened_vocab_ids.shape[0]
        unique_flattened_embeddings = np.zeros((unique_count, input_dim), dtype=np.float32)
        unique_flattened_count = np.zeros(unique_count, dtype=np.int32)

        # Use np.add.at to accumulate sums based on inverse indices
        np.add.at(unique_flattened_embeddings, inverse_indices, flattened_embeddings)
        np.add.at(unique_flattened_count, inverse_indices, 1)

        # Compute averages
        unique_flattened_embeddings /= unique_flattened_count[:, None]

        return unique_flattened_vocab_ids.astype(np.int32), unique_flattened_embeddings.astype(
            np.float32
        )

    def forward(
        self, vocab_ids: NumpyArray, embeddings: NumpyArray
    ) -> tuple[NumpyArray, NumpyArray]:
        """
        Args:
            vocab_ids: (batch_size, seq_len) int array
            embeddings: (batch_size, seq_len, input_dim) float array

        Returns:
            unique_flattened_vocab_ids_and_batch_ids: (total_unique, 2)
            unique_flattened_encoded: (total_unique, output_dim)
        """
        # Average embeddings for duplicate vocab_ids
        unique_flattened_vocab_ids_and_batch_ids, unique_flattened_embeddings = (
            self.avg_by_vocab_ids(vocab_ids, embeddings)
        )

        # Select the encoder weights for each unique vocab_id
        unique_flattened_vocab_ids = unique_flattened_vocab_ids_and_batch_ids[:, 0].astype(
            np.int32
        )

        # unique_encoder_weights: (total_unique, input_dim, output_dim)
        unique_encoder_weights = self.encoder_weights[unique_flattened_vocab_ids]

        # Compute linear transform: (total_unique, output_dim)
        # Using Einstein summation for matrix multiplication:
        # 'bi,bio->bo' means: for each "b" (batch element), multiply embeddings (b,i) by weights (b,i,o) -> (b,o)
        unique_flattened_encoded = np.einsum(
            "bi,bio->bo", unique_flattened_embeddings, unique_encoder_weights
        )

        # Apply Tanh activation and ensure float32 type
        unique_flattened_encoded = self.activation(unique_flattened_encoded).astype(np.float32)

        return unique_flattened_vocab_ids_and_batch_ids.astype(np.int32), unique_flattened_encoded



from typing import Any, Iterable, Type

import numpy as np
from numpy.typing import NDArray

from fastembed.common.types import NumpyArray
from fastembed.common.onnx_model import OnnxOutputContext
from fastembed.common.utils import mean_pooling
from fastembed.text.onnx_embedding import OnnxTextEmbedding, OnnxTextEmbeddingWorker
from fastembed.common.model_description import DenseModelDescription, ModelSource

supported_pooled_models: list[DenseModelDescription] = [
    DenseModelDescription(
        model="nomic-ai/nomic-embed-text-v1.5",
        dim=768,
        description=(
            "Text embeddings, Multimodal (text, image), English, 8192 input tokens truncation, "
            "Prefixes for queries/documents: necessary, 2024 year."
        ),
        license="apache-2.0",
        size_in_GB=0.52,
        sources=ModelSource(hf="nomic-ai/nomic-embed-text-v1.5"),
        model_file="onnx/model.onnx",
    ),
    DenseModelDescription(
        model="nomic-ai/nomic-embed-text-v1.5-Q",
        dim=768,
        description=(
            "Text embeddings, Multimodal (text, image), English, 8192 input tokens truncation, "
            "Prefixes for queries/documents: necessary, 2024 year."
        ),
        license="apache-2.0",
        size_in_GB=0.13,
        sources=ModelSource(hf="nomic-ai/nomic-embed-text-v1.5"),
        model_file="onnx/model_quantized.onnx",
    ),
    DenseModelDescription(
        model="nomic-ai/nomic-embed-text-v1",
        dim=768,
        description=(
            "Text embeddings, Multimodal (text, image), English, 8192 input tokens truncation, "
            "Prefixes for queries/documents: necessary, 2024 year."
        ),
        license="apache-2.0",
        size_in_GB=0.52,
        sources=ModelSource(hf="nomic-ai/nomic-embed-text-v1"),
        model_file="onnx/model.onnx",
    ),
    DenseModelDescription(
        model="sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2",
        dim=384,
        description=(
            "Text embeddings, Unimodal (text), Multilingual (~50 languages), 512 input tokens truncation, "
            "Prefixes for queries/documents: not necessary, 2019 year."
        ),
        license="apache-2.0",
        size_in_GB=0.22,
        sources=ModelSource(hf="qdrant/paraphrase-multilingual-MiniLM-L12-v2-onnx-Q"),
        model_file="model_optimized.onnx",
    ),
    DenseModelDescription(
        model="sentence-transformers/paraphrase-multilingual-mpnet-base-v2",
        dim=768,
        description=(
            "Text embeddings, Unimodal (text), Multilingual (~50 languages), 384 input tokens truncation, "
            "Prefixes for queries/documents: not necessary, 2021 year."
        ),
        license="apache-2.0",
        size_in_GB=1.00,
        sources=ModelSource(hf="xenova/paraphrase-multilingual-mpnet-base-v2"),
        model_file="onnx/model.onnx",
    ),
    DenseModelDescription(
        model="intfloat/multilingual-e5-large",
        dim=1024,
        description=(
            "Text embeddings, Unimodal (text), Multilingual (~100 languages), 512 input tokens truncation, "
            "Prefixes for queries/documents: necessary, 2024 year."
        ),
        license="mit",
        size_in_GB=2.24,
        sources=ModelSource(
            hf="qdrant/multilingual-e5-large-onnx",
            url="https://storage.googleapis.com/qdrant-fastembed/fast-multilingual-e5-large.tar.gz",
            _deprecated_tar_struct=True,
        ),
        model_file="model.onnx",
        additional_files=["model.onnx_data"],
    ),
]


class PooledEmbedding(OnnxTextEmbedding):
    @classmethod
    def _get_worker_class(cls) -> Type[OnnxTextEmbeddingWorker]:
        return PooledEmbeddingWorker

    @classmethod
    def mean_pooling(
        cls, model_output: NumpyArray, attention_mask: NDArray[np.int64]
    ) -> NumpyArray:
        return mean_pooling(model_output, attention_mask)

    @classmethod
    def _list_supported_models(cls) -> list[DenseModelDescription]:
        """Lists the supported models.

        Returns:
            list[DenseModelDescription]: A list of DenseModelDescription objects containing the model information.
        """
        return supported_pooled_models

    def _post_process_onnx_output(
        self, output: OnnxOutputContext, **kwargs: Any
    ) -> Iterable[NumpyArray]:
        if output.attention_mask is None:
            raise ValueError("attention_mask must be provided for document post-processing")

        embeddings = output.model_output
        attn_mask = output.attention_mask
        return self.mean_pooling(embeddings, attn_mask)


class PooledEmbeddingWorker(OnnxTextEmbeddingWorker):
    def init_embedding(
        self,
        model_name: str,
        cache_dir: str,
        **kwargs: Any,
    ) -> OnnxTextEmbedding:
        return PooledEmbedding(
            model_name=model_name,
            cache_dir=cache_dir,
            threads=1,
            **kwargs,
        )



from typing import Any, Iterable, Optional, Sequence, Type, Union

import numpy as np
from tokenizers import Encoding

from fastembed.common import OnnxProvider, ImageInput
from fastembed.common.onnx_model import OnnxOutputContext
from fastembed.common.types import NumpyArray
from fastembed.common.utils import define_cache_dir
from fastembed.late_interaction_multimodal.late_interaction_multimodal_embedding_base import (
    LateInteractionMultimodalEmbeddingBase,
)
from fastembed.late_interaction_multimodal.onnx_multimodal_model import (
    OnnxMultimodalModel,
    TextEmbeddingWorker,
    ImageEmbeddingWorker,
)
from fastembed.common.model_description import DenseModelDescription, ModelSource

supported_colpali_models: list[DenseModelDescription] = [
    DenseModelDescription(
        model="Qdrant/colpali-v1.3-fp16",
        dim=128,
        description="Text embeddings, Multimodal (text&image), English, 50 tokens query length truncation, 2024.",
        license="mit",
        size_in_GB=6.5,
        sources=ModelSource(hf="Qdrant/colpali-v1.3-fp16"),
        additional_files=["model.onnx_data"],
        model_file="model.onnx",
    ),
]


class ColPali(LateInteractionMultimodalEmbeddingBase, OnnxMultimodalModel[NumpyArray]):
    QUERY_PREFIX = "Query: "
    BOS_TOKEN = "<s>"
    PAD_TOKEN = "<pad>"
    QUERY_MARKER_TOKEN_ID = [2, 5098]
    IMAGE_PLACEHOLDER_SIZE = (3, 448, 448)
    EMPTY_TEXT_PLACEHOLDER = np.array(
        [257152] * 1024 + [2, 50721, 573, 2416, 235265, 108]
    )  # This is a tokenization of '<image>' * 1024 + '<bos>Describe the image.\n' line which is used as placeholder
    # while processing an image
    EVEN_ATTENTION_MASK = np.array([1] * 1030)

    def __init__(
        self,
        model_name: str,
        cache_dir: Optional[str] = None,
        threads: Optional[int] = None,
        providers: Optional[Sequence[OnnxProvider]] = None,
        cuda: bool = False,
        device_ids: Optional[list[int]] = None,
        lazy_load: bool = False,
        device_id: Optional[int] = None,
        specific_model_path: Optional[str] = None,
        **kwargs: Any,
    ):
        """
        Args:
            model_name (str): The name of the model to use.
            cache_dir (str, optional): The path to the cache directory.
                                       Can be set using the `FASTEMBED_CACHE_PATH` env variable.
                                       Defaults to `fastembed_cache` in the system's temp directory.
            threads (int, optional): The number of threads single onnxruntime session can use. Defaults to None.
            providers (Optional[Sequence[OnnxProvider]], optional): The list of onnxruntime providers to use.
                Mutually exclusive with the `cuda` and `device_ids` arguments. Defaults to None.
            cuda (bool, optional): Whether to use cuda for inference. Mutually exclusive with `providers`
                Defaults to False.
            device_ids (Optional[list[int]], optional): The list of device ids to use for data parallel processing in
                workers. Should be used with `cuda=True`, mutually exclusive with `providers`. Defaults to None.
            lazy_load (bool, optional): Whether to load the model during class initialization or on demand.
                Should be set to True when using multiple-gpu and parallel encoding. Defaults to False.
            device_id (Optional[int], optional): The device id to use for loading the model in the worker process.

        Raises:
            ValueError: If the model_name is not in the format <org>/<model> e.g. BAAI/bge-base-en.
        """

        super().__init__(model_name, cache_dir, threads, **kwargs)
        self.providers = providers
        self.lazy_load = lazy_load

        # List of device ids, that can be used for data parallel processing in workers
        self.device_ids = device_ids
        self.cuda = cuda

        # This device_id will be used if we need to load model in current process
        self.device_id: Optional[int] = None
        if device_id is not None:
            self.device_id = device_id
        elif self.device_ids is not None:
            self.device_id = self.device_ids[0]

        self.model_description = self._get_model_description(model_name)
        self.cache_dir = str(define_cache_dir(cache_dir))

        self._specific_model_path = specific_model_path
        self._model_dir = self.download_model(
            self.model_description,
            self.cache_dir,
            local_files_only=self._local_files_only,
            specific_model_path=self._specific_model_path,
        )
        self.mask_token_id = None
        self.pad_token_id = None

        if not self.lazy_load:
            self.load_onnx_model()

    @classmethod
    def _list_supported_models(cls) -> list[DenseModelDescription]:
        """Lists the supported models.

        Returns:
            list[DenseModelDescription]: A list of DenseModelDescription objects containing the model information.
        """
        return supported_colpali_models

    def load_onnx_model(self) -> None:
        self._load_onnx_model(
            model_dir=self._model_dir,
            model_file=self.model_description.model_file,
            threads=self.threads,
            providers=self.providers,
            cuda=self.cuda,
            device_id=self.device_id,
        )

    def _post_process_onnx_image_output(
        self,
        output: OnnxOutputContext,
    ) -> Iterable[NumpyArray]:
        """
        Post-process the ONNX model output to convert it into a usable format.

        Args:
            output (OnnxOutputContext): The raw output from the ONNX model.

        Returns:
            Iterable[NumpyArray]: Post-processed output as NumPy arrays.
        """
        assert self.model_description.dim is not None, "Model dim is not defined"
        return output.model_output.reshape(
            output.model_output.shape[0], -1, self.model_description.dim
        )

    def _post_process_onnx_text_output(
        self,
        output: OnnxOutputContext,
    ) -> Iterable[NumpyArray]:
        """
        Post-process the ONNX model output to convert it into a usable format.

        Args:
            output (OnnxOutputContext): The raw output from the ONNX model.

        Returns:
            Iterable[NumpyArray]: Post-processed output as NumPy arrays.
        """
        return output.model_output

    def tokenize(self, documents: list[str], **kwargs: Any) -> list[Encoding]:
        texts_query: list[str] = []
        for query in documents:
            query = self.BOS_TOKEN + self.QUERY_PREFIX + query + self.PAD_TOKEN * 10
            query += "\n"

            texts_query.append(query)
        encoded = self.tokenizer.encode_batch(texts_query)  # type: ignore[union-attr]
        return encoded

    def _preprocess_onnx_text_input(
        self, onnx_input: dict[str, NumpyArray], **kwargs: Any
    ) -> dict[str, NumpyArray]:
        onnx_input["input_ids"] = np.array(
            [
                self.QUERY_MARKER_TOKEN_ID + input_ids[2:].tolist()  # type: ignore[index]
                for input_ids in onnx_input["input_ids"]
            ]
        )
        empty_image_placeholder: NumpyArray = np.zeros(
            self.IMAGE_PLACEHOLDER_SIZE, dtype=np.float32
        )
        onnx_input["pixel_values"] = np.array(
            [empty_image_placeholder for _ in onnx_input["input_ids"]],
        )
        return onnx_input

    def _preprocess_onnx_image_input(
        self, onnx_input: dict[str, np.ndarray], **kwargs: Any
    ) -> dict[str, NumpyArray]:
        """
        Add placeholders for text input when processing image data for ONNX.
        Args:
            onnx_input (Dict[str, NumpyArray]): Preprocessed image inputs.
            **kwargs: Additional arguments.
        Returns:
            Dict[str, NumpyArray]: ONNX input with text placeholders.
        """
        onnx_input["input_ids"] = np.array(
            [self.EMPTY_TEXT_PLACEHOLDER for _ in onnx_input["pixel_values"]]
        )
        onnx_input["attention_mask"] = np.array(
            [self.EVEN_ATTENTION_MASK for _ in onnx_input["pixel_values"]]
        )
        return onnx_input

    def embed_text(
        self,
        documents: Union[str, Iterable[str]],
        batch_size: int = 256,
        parallel: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterable[NumpyArray]:
        """
        Encode a list of documents into list of embeddings.

        Args:
            documents: Iterator of documents or single document to embed
            batch_size: Batch size for encoding -- higher values will use more memory, but be faster
            parallel:
                If > 1, data-parallel encoding will be used, recommended for offline encoding of large datasets.
                If 0, use all available cores.
                If None, don't use data-parallel processing, use default onnxruntime threading instead.

        Returns:
            List of embeddings, one per document
        """
        yield from self._embed_documents(
            model_name=self.model_name,
            cache_dir=str(self.cache_dir),
            documents=documents,
            batch_size=batch_size,
            parallel=parallel,
            providers=self.providers,
            cuda=self.cuda,
            device_ids=self.device_ids,
            local_files_only=self._local_files_only,
            specific_model_path=self._specific_model_path,
            **kwargs,
        )

    def embed_image(
        self,
        images: Union[ImageInput, Iterable[ImageInput]],
        batch_size: int = 16,
        parallel: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterable[NumpyArray]:
        """
        Encode a list of images into list of embeddings.

        Args:
            images: Iterator of image paths or single image path to embed
            batch_size: Batch size for encoding -- higher values will use more memory, but be faster
            parallel:
                If > 1, data-parallel encoding will be used, recommended for offline encoding of large datasets.
                If 0, use all available cores.
                If None, don't use data-parallel processing, use default onnxruntime threading instead.

        Returns:
            List of embeddings, one per document
        """
        yield from self._embed_images(
            model_name=self.model_name,
            cache_dir=str(self.cache_dir),
            images=images,
            batch_size=batch_size,
            parallel=parallel,
            providers=self.providers,
            cuda=self.cuda,
            device_ids=self.device_ids,
            local_files_only=self._local_files_only,
            specific_model_path=self._specific_model_path,
            **kwargs,
        )

    @classmethod
    def _get_text_worker_class(cls) -> Type[TextEmbeddingWorker[NumpyArray]]:
        return ColPaliTextEmbeddingWorker

    @classmethod
    def _get_image_worker_class(cls) -> Type[ImageEmbeddingWorker[NumpyArray]]:
        return ColPaliImageEmbeddingWorker


class ColPaliTextEmbeddingWorker(TextEmbeddingWorker[NumpyArray]):
    def init_embedding(self, model_name: str, cache_dir: str, **kwargs: Any) -> ColPali:
        return ColPali(
            model_name=model_name,
            cache_dir=cache_dir,
            threads=1,
            **kwargs,
        )


class ColPaliImageEmbeddingWorker(ImageEmbeddingWorker[NumpyArray]):
    def init_embedding(self, model_name: str, cache_dir: str, **kwargs: Any) -> ColPali:
        return ColPali(
            model_name=model_name,
            cache_dir=cache_dir,
            threads=1,
            **kwargs,
        )



from typing import Any, Union, Optional

from PIL import Image

from fastembed.common.types import NumpyArray
from fastembed.image.transform.functional import (
    center_crop,
    convert_to_rgb,
    normalize,
    pil2ndarray,
    rescale,
    resize,
    pad2square,
)


class Transform:
    def __call__(self, images: list[Any]) -> Union[list[Image.Image], list[NumpyArray]]:
        raise NotImplementedError("Subclasses must implement this method")


class ConvertToRGB(Transform):
    def __call__(self, images: list[Image.Image]) -> list[Image.Image]:
        return [convert_to_rgb(image=image) for image in images]


class CenterCrop(Transform):
    def __init__(self, size: tuple[int, int]):
        self.size = size

    def __call__(self, images: list[Image.Image]) -> list[NumpyArray]:
        return [center_crop(image=image, size=self.size) for image in images]


class Normalize(Transform):
    def __init__(self, mean: Union[float, list[float]], std: Union[float, list[float]]):
        self.mean = mean
        self.std = std

    def __call__(self, images: list[NumpyArray]) -> list[NumpyArray]:
        return [normalize(image, mean=self.mean, std=self.std) for image in images]


class Resize(Transform):
    def __init__(
        self,
        size: Union[int, tuple[int, int]],
        resample: Image.Resampling = Image.Resampling.BICUBIC,
    ):
        self.size = size
        self.resample = resample

    def __call__(self, images: list[Image.Image]) -> list[Image.Image]:
        return [resize(image, size=self.size, resample=self.resample) for image in images]


class Rescale(Transform):
    def __init__(self, scale: float = 1 / 255):
        self.scale = scale

    def __call__(self, images: list[NumpyArray]) -> list[NumpyArray]:
        return [rescale(image, scale=self.scale) for image in images]


class PILtoNDarray(Transform):
    def __call__(self, images: list[Union[Image.Image, NumpyArray]]) -> list[NumpyArray]:
        return [pil2ndarray(image) for image in images]


class PadtoSquare(Transform):
    def __init__(
        self,
        size: int,
        fill_color: Union[str, int, tuple[int, ...]],
    ):
        self.size = size
        self.fill_color = fill_color

    def __call__(self, images: list[Image.Image]) -> list[Image.Image]:
        return [
            pad2square(image=image, size=self.size, fill_color=self.fill_color) for image in images
        ]


class Compose:
    def __init__(self, transforms: list[Transform]):
        self.transforms = transforms

    def __call__(
        self, images: Union[list[Image.Image], list[NumpyArray]]
    ) -> Union[list[NumpyArray], list[Image.Image]]:
        for transform in self.transforms:
            images = transform(images)
        return images

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> "Compose":
        """Creates processor from a config dict.
        Args:
            config (dict[str, Any]): Configuration dictionary.

                Valid keys:
                    - do_resize
                    - resize_mode
                    - size
                    - fill_color
                    - do_center_crop
                    - crop_size
                    - do_rescale
                    - rescale_factor
                    - do_normalize
                    - image_mean
                    - mean
                    - image_std
                    - std
                    - resample
                    - interpolation
                Valid size keys (nested):
                    - {"height", "width"}
                    - {"shortest_edge"}

        Returns:
            Compose: Image processor.
        """
        transforms: list[Transform] = []
        cls._get_convert_to_rgb(transforms, config)
        cls._get_resize(transforms, config)
        cls._get_pad2square(transforms, config)
        cls._get_center_crop(transforms, config)
        cls._get_pil2ndarray(transforms, config)
        cls._get_rescale(transforms, config)
        cls._get_normalize(transforms, config)
        return cls(transforms=transforms)

    @staticmethod
    def _get_convert_to_rgb(transforms: list[Transform], config: dict[str, Any]) -> None:
        transforms.append(ConvertToRGB())

    @classmethod
    def _get_resize(cls, transforms: list[Transform], config: dict[str, Any]) -> None:
        mode = config.get("image_processor_type", "CLIPImageProcessor")
        if mode in ("CLIPImageProcessor", "SiglipImageProcessor"):
            if config.get("do_resize", False):
                size = config["size"]
                if "shortest_edge" in size:
                    size = size["shortest_edge"]
                elif "height" in size and "width" in size:
                    size = (size["height"], size["width"])
                else:
                    raise ValueError(
                        "Size must contain either 'shortest_edge' or 'height' and 'width'."
                    )
                transforms.append(
                    Resize(
                        size=size,
                        resample=config.get("resample", Image.Resampling.BICUBIC),
                    )
                )
        elif mode == "ConvNextFeatureExtractor":
            if "size" in config and "shortest_edge" not in config["size"]:
                raise ValueError(
                    f"Size dictionary must contain 'shortest_edge' key. Got {config['size'].keys()}"
                )
            shortest_edge = config["size"]["shortest_edge"]
            crop_pct = config.get("crop_pct", 0.875)
            if shortest_edge < 384:
                # maintain same ratio, resizing shortest edge to shortest_edge/crop_pct
                resize_shortest_edge = int(shortest_edge / crop_pct)
                transforms.append(
                    Resize(
                        size=resize_shortest_edge,
                        resample=config.get("resample", Image.Resampling.BICUBIC),
                    )
                )
                transforms.append(CenterCrop(size=(shortest_edge, shortest_edge)))
            else:
                transforms.append(
                    Resize(
                        size=(shortest_edge, shortest_edge),
                        resample=config.get("resample", Image.Resampling.BICUBIC),
                    )
                )
        elif mode == "JinaCLIPImageProcessor":
            interpolation = config.get("interpolation")
            if isinstance(interpolation, str):
                resample = cls._interpolation_resolver(interpolation)
            else:
                resample = interpolation or Image.Resampling.BICUBIC

            if "size" in config:
                resize_mode = config.get("resize_mode", "shortest")
                if resize_mode == "shortest":
                    transforms.append(
                        Resize(
                            size=config["size"],
                            resample=resample,
                        )
                    )
        else:
            raise ValueError(f"Preprocessor {mode} is not supported")

    @staticmethod
    def _get_center_crop(transforms: list[Transform], config: dict[str, Any]) -> None:
        mode = config.get("image_processor_type", "CLIPImageProcessor")
        if mode in ("CLIPImageProcessor", "SiglipImageProcessor"):
            if config.get("do_center_crop", False):
                crop_size_raw = config["crop_size"]
                crop_size: tuple[int, int]
                if isinstance(crop_size_raw, int):
                    crop_size = (crop_size_raw, crop_size_raw)
                elif isinstance(crop_size_raw, dict):
                    crop_size = (crop_size_raw["height"], crop_size_raw["width"])
                else:
                    raise ValueError(f"Invalid crop size: {crop_size_raw}")
                transforms.append(CenterCrop(size=crop_size))
        elif mode == "ConvNextFeatureExtractor":
            pass
        elif mode == "JinaCLIPImageProcessor":
            pass
        else:
            raise ValueError(f"Preprocessor {mode} is not supported")

    @staticmethod
    def _get_pil2ndarray(transforms: list[Transform], config: dict[str, Any]) -> None:
        transforms.append(PILtoNDarray())

    @staticmethod
    def _get_rescale(transforms: list[Transform], config: dict[str, Any]) -> None:
        if config.get("do_rescale", True):
            rescale_factor = config.get("rescale_factor", 1 / 255)
            transforms.append(Rescale(scale=rescale_factor))

    @staticmethod
    def _get_normalize(transforms: list[Transform], config: dict[str, Any]) -> None:
        if config.get("do_normalize", False):
            transforms.append(Normalize(mean=config["image_mean"], std=config["image_std"]))
        elif "mean" in config and "std" in config:
            transforms.append(Normalize(mean=config["mean"], std=config["std"]))

    @staticmethod
    def _get_pad2square(transforms: list[Transform], config: dict[str, Any]) -> None:
        mode = config.get("image_processor_type", "CLIPImageProcessor")
        if mode == "CLIPImageProcessor":
            pass
        elif mode == "ConvNextFeatureExtractor":
            pass
        elif mode == "JinaCLIPImageProcessor":
            transforms.append(
                PadtoSquare(
                    size=config["size"],
                    fill_color=config.get("fill_color", 0),
                )
            )

    @staticmethod
    def _interpolation_resolver(resample: Optional[str] = None) -> Image.Resampling:
        interpolation_map = {
            "nearest": Image.Resampling.NEAREST,
            "lanczos": Image.Resampling.LANCZOS,
            "bilinear": Image.Resampling.BILINEAR,
            "bicubic": Image.Resampling.BICUBIC,
            "box": Image.Resampling.BOX,
            "hamming": Image.Resampling.HAMMING,
        }

        if resample and (method := interpolation_map.get(resample.lower())):
            return method

        raise ValueError(f"Unknown interpolation method: {resample}")



from typing import Any, Iterable, Optional, Sequence, Type, Union
from dataclasses import asdict

from fastembed.common import OnnxProvider
from fastembed.sparse.bm25 import Bm25
from fastembed.sparse.bm42 import Bm42
from fastembed.sparse.minicoil import MiniCOIL
from fastembed.sparse.sparse_embedding_base import (
    SparseEmbedding,
    SparseTextEmbeddingBase,
)
from fastembed.sparse.splade_pp import SpladePP
import warnings
from fastembed.common.model_description import SparseModelDescription


class SparseTextEmbedding(SparseTextEmbeddingBase):
    EMBEDDINGS_REGISTRY: list[Type[SparseTextEmbeddingBase]] = [SpladePP, Bm42, Bm25, MiniCOIL]

    @classmethod
    def list_supported_models(cls) -> list[dict[str, Any]]:
        """
        Lists the supported models.

        Returns:
            list[dict[str, Any]]: A list of dictionaries containing the model information.

            Example:
                ```
                [
                    {
                        "model": "prithvida/SPLADE_PP_en_v1",
                        "vocab_size": 30522,
                        "description": "Independent Implementation of SPLADE++ Model for English",
                        "license": "apache-2.0",
                        "size_in_GB": 0.532,
                        "sources": {
                            "hf": "qdrant/SPLADE_PP_en_v1",
                        },
                    }
                ]
                ```
        """
        return [asdict(model) for model in cls._list_supported_models()]

    @classmethod
    def _list_supported_models(cls) -> list[SparseModelDescription]:
        result: list[SparseModelDescription] = []
        for embedding in cls.EMBEDDINGS_REGISTRY:
            result.extend(embedding._list_supported_models())
        return result

    def __init__(
        self,
        model_name: str,
        cache_dir: Optional[str] = None,
        threads: Optional[int] = None,
        providers: Optional[Sequence[OnnxProvider]] = None,
        cuda: bool = False,
        device_ids: Optional[list[int]] = None,
        lazy_load: bool = False,
        **kwargs: Any,
    ):
        super().__init__(model_name, cache_dir, threads, **kwargs)
        if model_name.lower() == "prithvida/Splade_PP_en_v1".lower():
            warnings.warn(
                "The right spelling is prithivida/Splade_PP_en_v1. "
                "Support of this name will be removed soon, please fix the model_name",
                DeprecationWarning,
                stacklevel=2,
            )
            model_name = "prithivida/Splade_PP_en_v1"

        for EMBEDDING_MODEL_TYPE in self.EMBEDDINGS_REGISTRY:
            supported_models = EMBEDDING_MODEL_TYPE._list_supported_models()
            if any(model_name.lower() == model.model.lower() for model in supported_models):
                self.model = EMBEDDING_MODEL_TYPE(
                    model_name,
                    cache_dir,
                    threads=threads,
                    providers=providers,
                    cuda=cuda,
                    device_ids=device_ids,
                    lazy_load=lazy_load,
                    **kwargs,
                )
                return

        raise ValueError(
            f"Model {model_name} is not supported in SparseTextEmbedding."
            "Please check the supported models using `SparseTextEmbedding.list_supported_models()`"
        )

    def embed(
        self,
        documents: Union[str, Iterable[str]],
        batch_size: int = 256,
        parallel: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterable[SparseEmbedding]:
        """
        Encode a list of documents into list of embeddings.
        We use mean pooling with attention so that the model can handle variable-length inputs.

        Args:
            documents: Iterator of documents or single document to embed
            batch_size: Batch size for encoding -- higher values will use more memory, but be faster
            parallel:
                If > 1, data-parallel encoding will be used, recommended for offline encoding of large datasets.
                If 0, use all available cores.
                If None, don't use data-parallel processing, use default onnxruntime threading instead.

        Returns:
            List of embeddings, one per document
        """
        yield from self.model.embed(documents, batch_size, parallel, **kwargs)

    def query_embed(
        self, query: Union[str, Iterable[str]], **kwargs: Any
    ) -> Iterable[SparseEmbedding]:
        """
        Embeds queries

        Args:
            query (Union[str, Iterable[str]]): The query to embed, or an iterable e.g. list of queries.

        Returns:
            Iterable[SparseEmbedding]: The sparse embeddings.
        """
        yield from self.model.query_embed(query, **kwargs)



from pathlib import Path

from typing import Any, Optional, Sequence, Iterable, Union, Type

import numpy as np
from numpy.typing import NDArray
from py_rust_stemmers import SnowballStemmer
from tokenizers import Tokenizer

from fastembed.common.model_description import SparseModelDescription, ModelSource
from fastembed.common.onnx_model import OnnxOutputContext
from fastembed.common import OnnxProvider
from fastembed.common.utils import define_cache_dir
from fastembed.sparse.sparse_embedding_base import (
    SparseEmbedding,
    SparseTextEmbeddingBase,
)
from fastembed.sparse.utils.minicoil_encoder import Encoder
from fastembed.sparse.utils.sparse_vectors_converter import SparseVectorConverter, WordEmbedding
from fastembed.sparse.utils.vocab_resolver import VocabResolver, VocabTokenizer
from fastembed.text.onnx_text_model import OnnxTextModel, TextEmbeddingWorker


MINICOIL_MODEL_FILE = "minicoil.triplet.model.npy"
MINICOIL_VOCAB_FILE = "minicoil.triplet.model.vocab"
STOPWORDS_FILE = "stopwords.txt"


supported_minicoil_models: list[SparseModelDescription] = [
    SparseModelDescription(
        model="Qdrant/minicoil-v1",
        vocab_size=19125,
        description="Sparse embedding model, that resolves semantic meaning of the words, "
        "while keeping exact keyword match behavior. "
        "Based on jinaai/jina-embeddings-v2-small-en-tokens",
        license="apache-2.0",
        size_in_GB=0.09,
        sources=ModelSource(hf="Qdrant/minicoil-v1"),
        model_file="onnx/model.onnx",
        additional_files=[
            STOPWORDS_FILE,
            MINICOIL_MODEL_FILE,
            MINICOIL_VOCAB_FILE,
        ],
        requires_idf=True,
    ),
]

MODEL_TO_LANGUAGE = {
    "Qdrant/minicoil-v1": "english",
}


class MiniCOIL(SparseTextEmbeddingBase, OnnxTextModel[SparseEmbedding]):
    """
        MiniCOIL is a sparse embedding model, that resolves semantic meaning of the words,
        while keeping exact keyword match behavior.

        Each vocabulary token is converted into 4d component of a sparse vector, which is then weighted by the token frequency in the corpus.
        If the token is not found in the corpus, it is treated exactly like in BM25.
    `
        The model is based on `jinaai/jina-embeddings-v2-small-en-tokens`
    """

    def __init__(
        self,
        model_name: str,
        cache_dir: Optional[str] = None,
        threads: Optional[int] = None,
        providers: Optional[Sequence[OnnxProvider]] = None,
        k: float = 1.2,
        b: float = 0.75,
        avg_len: float = 150.0,
        cuda: bool = False,
        device_ids: Optional[list[int]] = None,
        lazy_load: bool = False,
        device_id: Optional[int] = None,
        specific_model_path: Optional[str] = None,
        **kwargs: Any,
    ):
        """
        Args:
            model_name (str): The name of the model to use.
            cache_dir (str, optional): The path to the cache directory.
                                       Can be set using the `FASTEMBED_CACHE_PATH` env variable.
                                       Defaults to `fastembed_cache` in the system's temp directory.
            threads (int, optional): The number of threads single onnxruntime session can use. Defaults to None.
            providers (Optional[Sequence[OnnxProvider]], optional): The providers to use for onnxruntime.
            k (float, optional): The k parameter in the BM25 formula. Defines the saturation of the term frequency.
                I.e. defines how fast the moment when additional terms stop to increase the score. Defaults to 1.2.
            b (float, optional): The b parameter in the BM25 formula. Defines the importance of the document length.
                Defaults to 0.75.
            avg_len (float, optional): The average length of the documents in the corpus. Defaults to 150.0.
            cuda (bool, optional): Whether to use cuda for inference. Mutually exclusive with `providers`
                Defaults to False.
            device_ids (Optional[list[int]], optional): The list of device ids to use for data parallel processing in
                workers. Should be used with `cuda=True`, mutually exclusive with `providers`. Defaults to None.
            lazy_load (bool, optional): Whether to load the model during class initialization or on demand.
                Should be set to True when using multiple-gpu and parallel encoding. Defaults to False.
            device_id (Optional[int], optional): The device id to use for loading the model in the worker process.
            specific_model_path (Optional[str], optional): The specific path to the onnx model dir if it should be imported from somewhere else

        Raises:
            ValueError: If the model_name is not in the format <org>/<model> e.g. BAAI/bge-base-en.
        """

        super().__init__(model_name, cache_dir, threads, **kwargs)
        self.providers = providers
        self.lazy_load = lazy_load
        self.device_ids = device_ids
        self.cuda = cuda
        self.device_id = device_id
        self.k = k
        self.b = b
        self.avg_len = avg_len

        # Initialize class attributes
        self.tokenizer: Optional[Tokenizer] = None
        self.invert_vocab: dict[int, str] = {}
        self.special_tokens: set[str] = set()
        self.special_tokens_ids: set[int] = set()
        self.stopwords: set[str] = set()
        self.vocab_resolver: Optional[VocabResolver] = None
        self.encoder: Optional[Encoder] = None
        self.output_dim: Optional[int] = None
        self.sparse_vector_converter: Optional[SparseVectorConverter] = None

        self.model_description = self._get_model_description(model_name)
        self.cache_dir = str(define_cache_dir(cache_dir))
        self._specific_model_path = specific_model_path
        self._model_dir = self.download_model(
            self.model_description,
            self.cache_dir,
            local_files_only=self._local_files_only,
            specific_model_path=self._specific_model_path,
        )

        if not self.lazy_load:
            self.load_onnx_model()

    def load_onnx_model(self) -> None:
        self._load_onnx_model(
            model_dir=self._model_dir,
            model_file=self.model_description.model_file,
            threads=self.threads,
            providers=self.providers,
            cuda=self.cuda,
            device_id=self.device_id,
        )

        assert self.tokenizer is not None

        for token, idx in self.tokenizer.get_vocab().items():  # type: ignore[union-attr]
            self.invert_vocab[idx] = token
        self.special_tokens = set(self.special_token_to_id.keys())
        self.special_tokens_ids = set(self.special_token_to_id.values())
        self.stopwords = set(self._load_stopwords(self._model_dir))

        stemmer = SnowballStemmer(MODEL_TO_LANGUAGE[self.model_name])

        self.vocab_resolver = VocabResolver(
            tokenizer=VocabTokenizer(self.tokenizer),
            stopwords=self.stopwords,
            stemmer=stemmer,
        )
        self.vocab_resolver.load_json_vocab(str(self._model_dir / MINICOIL_VOCAB_FILE))

        weights = np.load(str(self._model_dir / MINICOIL_MODEL_FILE), mmap_mode="r")
        self.encoder = Encoder(weights)
        self.output_dim = self.encoder.output_dim

        self.sparse_vector_converter = SparseVectorConverter(
            stopwords=self.stopwords,
            stemmer=stemmer,
            k=self.k,
            b=self.b,
            avg_len=self.avg_len,
        )

    def embed(
        self,
        documents: Union[str, Iterable[str]],
        batch_size: int = 256,
        parallel: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterable[SparseEmbedding]:
        """
        Encode a list of documents into list of embeddings.
        We use mean pooling with attention so that the model can handle variable-length inputs.

        Args:
            documents: Iterator of documents or single document to embed
            batch_size: Batch size for encoding -- higher values will use more memory, but be faster
            parallel:
                If > 1, data-parallel encoding will be used, recommended for offline encoding of large datasets.
                If 0, use all available cores.
                If None, don't use data-parallel processing, use default onnxruntime threading instead.

        Returns:
            List of embeddings, one per document
        """
        yield from self._embed_documents(
            model_name=self.model_name,
            cache_dir=str(self.cache_dir),
            documents=documents,
            batch_size=batch_size,
            parallel=parallel,
            providers=self.providers,
            cuda=self.cuda,
            device_ids=self.device_ids,
            k=self.k,
            b=self.b,
            avg_len=self.avg_len,
            is_query=False,
            local_files_only=self._local_files_only,
            specific_model_path=self._specific_model_path,
            **kwargs,
        )

    def query_embed(
        self, query: Union[str, Iterable[str]], **kwargs: Any
    ) -> Iterable[SparseEmbedding]:
        """
        Encode a list of queries into list of embeddings.
        """
        yield from self._embed_documents(
            model_name=self.model_name,
            cache_dir=str(self.cache_dir),
            documents=query,
            providers=self.providers,
            cuda=self.cuda,
            device_ids=self.device_ids,
            k=self.k,
            b=self.b,
            avg_len=self.avg_len,
            is_query=True,
            local_files_only=self._local_files_only,
            specific_model_path=self._specific_model_path,
            **kwargs,
        )

    @classmethod
    def _load_stopwords(cls, model_dir: Path) -> list[str]:
        stopwords_path = model_dir / STOPWORDS_FILE
        if not stopwords_path.exists():
            return []

        with open(stopwords_path, "r") as f:
            return f.read().splitlines()

    @classmethod
    def _list_supported_models(cls) -> list[SparseModelDescription]:
        """Lists the supported models.

        Returns:
            list[SparseModelDescription]: A list of SparseModelDescription objects containing the model information.
        """
        return supported_minicoil_models

    def _post_process_onnx_output(
        self, output: OnnxOutputContext, is_query: bool = False, **kwargs: Any
    ) -> Iterable[SparseEmbedding]:
        if output.input_ids is None:
            raise ValueError("input_ids must be provided for document post-processing")

        assert self.vocab_resolver is not None
        assert self.encoder is not None
        assert self.sparse_vector_converter is not None

        # Size: (batch_size, sequence_length, hidden_size)
        embeddings = output.model_output
        # Size: (batch_size, sequence_length)
        assert output.attention_mask is not None
        masks = output.attention_mask

        vocab_size = self.vocab_resolver.vocab_size()
        embedding_size = self.encoder.output_dim

        # For each document we only select those embeddings that are not masked out

        for i in range(embeddings.shape[0]):
            # Size: (sequence_length, hidden_size)
            token_embeddings = embeddings[i, masks[i] == 1]

            # Size: (sequence_length)
            token_ids: NDArray[np.int64] = output.input_ids[i, masks[i] == 1]

            word_ids_array, counts, oov, forms = self.vocab_resolver.resolve_tokens(token_ids)

            # Size: (1, words)
            word_ids_array_expanded: NDArray[np.int64] = np.expand_dims(word_ids_array, axis=0)

            # Size: (1, words, embedding_size)
            token_embeddings_array: NDArray[np.float32] = np.expand_dims(token_embeddings, axis=0)

            assert word_ids_array_expanded.shape[1] == token_embeddings_array.shape[1]

            # Size of word_ids_mapping: (unique_words, 2) - [vocab_id, batch_id]
            # Size of embeddings: (unique_words, embedding_size)
            ids_mapping, minicoil_embeddings = self.encoder.forward(
                word_ids_array_expanded, token_embeddings_array
            )

            # Size of counts: (unique_words)
            words_ids: list[int] = ids_mapping[:, 0].tolist()  # type: ignore[assignment]

            sentence_result: dict[str, WordEmbedding] = {}

            words = [self.vocab_resolver.lookup_word(word_id) for word_id in words_ids]

            for word, word_id, emb in zip(words, words_ids, minicoil_embeddings.tolist()):  # type: ignore[arg-type]
                if word_id == 0:
                    continue

                sentence_result[word] = WordEmbedding(
                    word=word,
                    forms=forms[word],
                    count=int(counts[word_id]),
                    word_id=int(word_id),
                    embedding=emb,  # type: ignore[arg-type]
                )

            for oov_word, count in oov.items():
                # {
                #     "word": oov_word,
                #     "forms": [oov_word],
                #     "count": int(count),
                #     "word_id": -1,
                #     "embedding": [1]
                # }
                sentence_result[oov_word] = WordEmbedding(
                    word=oov_word, forms=[oov_word], count=int(count), word_id=-1, embedding=[1]
                )

            if not is_query:
                yield self.sparse_vector_converter.embedding_to_vector(
                    sentence_result, vocab_size=vocab_size, embedding_size=embedding_size
                )
            else:
                yield self.sparse_vector_converter.embedding_to_vector_query(
                    sentence_result, vocab_size=vocab_size, embedding_size=embedding_size
                )

    @classmethod
    def _get_worker_class(cls) -> Type["MiniCoilTextEmbeddingWorker"]:
        return MiniCoilTextEmbeddingWorker


class MiniCoilTextEmbeddingWorker(TextEmbeddingWorker[SparseEmbedding]):
    def init_embedding(self, model_name: str, cache_dir: str, **kwargs: Any) -> MiniCOIL:
        return MiniCOIL(
            model_name=model_name,
            cache_dir=cache_dir,
            threads=1,
            **kwargs,
        )



# This code is a modified copy of the `NLTKWordTokenizer` class from `NLTK` library.

import re


class SimpleTokenizer:
    @staticmethod
    def tokenize(text: str) -> list[str]:
        text = re.sub(r"[^\w]", " ", text.lower())
        text = re.sub(r"\s+", " ", text)

        return text.strip().split()


class WordTokenizer:
    """The tokenizer is "destructive" such that the regexes applied will munge the
    input string to a state beyond re-construction.
    """

    # Starting quotes.
    STARTING_QUOTES = [
        (re.compile("([«“‘„]|[`]+)", re.U), r" \1 "),
        (re.compile(r"^\""), r"``"),
        (re.compile(r"(``)"), r" \1 "),
        (re.compile(r"([ \(\[{<])(\"|\'{2})"), r"\1 `` "),
        (re.compile(r"(?i)(\')(?!re|ve|ll|m|t|s|d|n)(\w)\b", re.U), r"\1 \2"),
    ]

    # Ending quotes.
    ENDING_QUOTES = [
        (re.compile("([»”’])", re.U), r" \1 "),
        (re.compile(r"''"), " '' "),
        (re.compile(r'"'), " '' "),
        (re.compile(r"([^' ])('[sS]|'[mM]|'[dD]|') "), r"\1 \2 "),
        (re.compile(r"([^' ])('ll|'LL|'re|'RE|'ve|'VE|n't|N'T) "), r"\1 \2 "),
    ]

    # Punctuation.
    PUNCTUATION = [
        (re.compile(r'([^\.])(\.)([\]\)}>"\'' "»”’ " r"]*)\s*$", re.U), r"\1 \2 \3 "),
        (re.compile(r"([:,])([^\d])"), r" \1 \2"),
        (re.compile(r"([:,])$"), r" \1 "),
        (
            re.compile(r"\.{2,}", re.U),
            r" \g<0> ",
        ),
        (re.compile(r"[;@#$%&]"), r" \g<0> "),
        (
            re.compile(r'([^\.])(\.)([\]\)}>"\']*)\s*$'),
            r"\1 \2\3 ",
        ),  # Handles the final period.
        (re.compile(r"[?!]"), r" \g<0> "),
        (re.compile(r"([^'])' "), r"\1 ' "),
        (
            re.compile(r"[*]", re.U),
            r" \g<0> ",
        ),
    ]

    # Pads parentheses
    PARENS_BRACKETS = (re.compile(r"[\]\[\(\)\{\}\<\>]"), r" \g<0> ")
    DOUBLE_DASHES = (re.compile(r"--"), r" -- ")

    # List of contractions adapted from Robert MacIntyre's tokenizer.
    CONTRACTIONS2 = [
        re.compile(pattern)
        for pattern in (
            r"(?i)\b(can)(?#X)(not)\b",
            r"(?i)\b(d)(?#X)('ye)\b",
            r"(?i)\b(gim)(?#X)(me)\b",
            r"(?i)\b(gon)(?#X)(na)\b",
            r"(?i)\b(got)(?#X)(ta)\b",
            r"(?i)\b(lem)(?#X)(me)\b",
            r"(?i)\b(more)(?#X)('n)\b",
            r"(?i)\b(wan)(?#X)(na)(?=\s)",
        )
    ]
    CONTRACTIONS3 = [
        re.compile(pattern) for pattern in (r"(?i) ('t)(?#X)(is)\b", r"(?i) ('t)(?#X)(was)\b")
    ]

    @classmethod
    def tokenize(cls, text: str) -> list[str]:
        """Return a tokenized copy of `text`.

        >>> s = '''Good muffins cost $3.88 (roughly 3,36 euros)\nin New York.'''
        >>> WordTokenizer().tokenize(s)
        ['Good', 'muffins', 'cost', '$', '3.88', '(', 'roughly', '3,36', 'euros', ')', 'in', 'New', 'York', '.']

        Args:
            text: The text to be tokenized.

        Returns:
            A list of tokens.
        """
        for regexp, substitution in cls.STARTING_QUOTES:
            text = regexp.sub(substitution, text)

        for regexp, substitution in cls.PUNCTUATION:
            text = regexp.sub(substitution, text)

        # Handles parentheses.
        regexp, substitution = cls.PARENS_BRACKETS
        text = regexp.sub(substitution, text)

        # Handles double dash.
        regexp, substitution = cls.DOUBLE_DASHES
        text = regexp.sub(substitution, text)

        # add extra space to make things easier
        text = " " + text + " "

        for regexp, substitution in cls.ENDING_QUOTES:
            text = regexp.sub(substitution, text)

        for regexp in cls.CONTRACTIONS2:
            text = regexp.sub(r" \1 \2 ", text)
        for regexp in cls.CONTRACTIONS3:
            text = regexp.sub(r" \1 \2 ", text)
        return text.split()
